"""
Models to express `Affect.versions` fields
"""
from __future__ import annotations

from typing import Annotated, ClassVar

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from pydantic import BaseModel, Field

from hoppr.models.base import CycloneDXBaseModel, UniqueIDMap


class AffectVersion(BaseModel):
    """
    AffectVersion data model representing properties of `versions` items
    """

    version: cdx.Version | None = Field(default=None)
    range: cdx.Range | None = Field(default=None)
    status: cdx.AffectedStatus | None = Field(
        default=cdx.AffectedStatus.affected,
        description="The vulnerability status for the version or range of versions.",
    )


class AffectVersionVersionRequired(AffectVersion):
    """
    Affect.versions item model with required `version` field
    """

    version: cdx.Version


class AffectVersionRangeRequired(AffectVersion):
    """
    Affect.versions item model with required `range` field
    """

    range: cdx.Range


AffectVersions = list[AffectVersionVersionRequired | AffectVersionRangeRequired] | None


class Affect(CycloneDXBaseModel, cdx.Affect):
    """
    Affect data model derived from CycloneDXBaseModel with overridden `versions` field
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Affect model"

    versions: Annotated[
        AffectVersions,
        Field(
            description="Zero or more individual versions or range of versions.",
            title="Versions",
        ),
    ] = None

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}



# pylint: disable=all
from collections.abc import Callable, MutableMapping
from itertools import count
from logging import Logger, LoggerAdapter
from threading import _RLock as RLock

from typing_extensions import override

def locked(func: Callable) -> Callable: ...

class HopprLogger(LoggerAdapter):
    id_iter: count[int]
    extra: MutableMapping[str, object]
    logger: Logger
    filename: str
    flush_immed: bool
    instance_id: int
    lock: RLock | None
    def __init__(
        self,
        name: str | None = ...,
        filename: str = ...,
        level: int = ...,
        lock: RLock | None = ...,
        flush_immed: bool = ...,
    ) -> None: ...
    def clear_targets(self) -> None: ...
    def close(self) -> None: ...
    def flush(self) -> None: ...
    def is_verbose(self) -> bool: ...
    @override
    def log(  # type: ignore[override]
        self,
        level: int,
        msg: object,
        *args: object,
        indent_level: int = 0,
        **kwargs: object,
    ) -> None: ...
    def process(self, msg: str, kwargs: MutableMapping[str, object]) -> tuple[str, MutableMapping[str, object]]: ...
    @override
    def debug(self, msg: str, *args, indent_level: int = 0, **kwargs) -> None:  # type: ignore[override]
        """
        Wrapper function for debug messages
        """
    @override
    def info(self, msg: str, *args, indent_level: int = 0, **kwargs) -> None:  # type: ignore[override]
        """
        Wrapper function for info messages
        """
    @override
    def warning(self, msg: str, *args, indent_level: int = 0, **kwargs) -> None:  # type: ignore[override]
        """
        Wrapper function for warn messages
        """
    @override
    def error(self, msg: str, *args, indent_level: int = 0, **kwargs) -> None:  # type: ignore[override]
        """
        Wrapper function for error messages
        """
    @override
    def critical(self, msg: str, *args, indent_level: int = 0, **kwargs) -> None:  # type: ignore[override]
        """
        Wrapper function for critical messages
        """



"""
`bundle` subcommand for `hoptctl`
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional

from rich.console import Group
from rich.progress import MofNCompleteColumn, Progress
from rich.table import Table
from rich.tree import Tree
from typer import Argument, Context, Option, Typer

import hoppr.utils

from hoppr import main, processor
from hoppr.cli.layout import HopprBasePanel, HopprJobsPanel, HopprLayout, HopprSpinnerColumn
from hoppr.cli.options import basic_term_option, experimental_option, strict_repos_option, verbose_option

if TYPE_CHECKING:
    from rich.console import Console, ConsoleOptions, RenderResult

app = Typer(
    context_settings={"allow_interspersed_args": True, "help_option_names": ["-h", "--help"]},
    help="Run the stages specified in the transfer config file on the content specified in the manifest",
    invoke_without_command=True,
    no_args_is_help=True,
    rich_markup_mode="markdown",
    subcommand_metavar="",
)


class HopprBundleJobsPanel(HopprJobsPanel):
    """
    Customized Rich Progress bar Panel
    """

    progress_bar = Progress(
        "{task.description}",
        MofNCompleteColumn(),
        HopprSpinnerColumn(),
        expand=True,
    )

    def __init__(self) -> None:
        super().__init__()
        self.renderable = self.progress_bar
        self.title = "[bold blue]Components"


class HopprBundleLayout(HopprLayout):
    """
    Layout of the `hopctl bundle` console application
    """

    name: str = "root"
    jobs_panel = HopprBundleJobsPanel()


class HopprBundleSummary(HopprBasePanel):
    """
    `hopctl bundle` results summary
    """

    stage_results_map: dict[str, Table] = {}

    def __init__(self, *args, title: str | None = "Summary", **kwargs) -> None:
        self.summary_group = Group()
        self.failure_table = Table("Plugin", "Component", "Details", box=None, pad_edge=False, show_edge=False)

        self.total_success_count = 0
        self.total_failure_count = 0

        super().__init__(self.summary_group, title, *args, **kwargs)

    def add_failure(self, plugin_name: str, comp_str: str | None, message: str) -> None:
        """
        Add an entry to the failures table

        Args:
            plugin_name (str): Name of the plugin that failed
            comp_str (str | None): Name of the failed component
            message (str): Result message
        """
        self.failure_table.add_row(plugin_name, comp_str or "", message)
        self.total_failure_count += 1
        self.total_success_count -= 1

    def add_method_result(self, stage_name: str, method_name: str, result_count: int, failure_count: int) -> None:
        """
        Add an entry to the results table for a plugin method

        Args:
            stage_name (str): Name of the stage
            method_name (str): Name of the method
            result_count (int): Total number of results
            failure_count (int): Number of failed results
        """
        results_table = self.stage_results_map[stage_name]
        results_table.add_row(method_name, str(result_count - failure_count), str(failure_count))

    def add_stage_result(self, stage_name: str) -> Tree:
        """
        Add a results table for the specified stage to the panel

        Args:
            stage_name (str): Name of the stage

        Returns:
            Tree: Object containing nested results table
        """
        self.stage_results_map[stage_name] = Table(
            "Step",
            "# Success",
            "# Fail",
            box=None,
            pad_edge=False,
            show_edge=False,
        )

        stage_summary = Tree(label=f"Stage: {stage_name}", guide_style="conceal")
        stage_summary.add(self.stage_results_map[stage_name])

        return stage_summary

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        if self.total_failure_count > 0:
            self.summary_group.renderables.append(self.failure_table)
            self.failure_table.add_row()

        self.summary_group.renderables.append(
            f"GRAND TOTAL: {self.total_success_count} jobs succeeded, {self.total_failure_count} failed"
        )

        if hoppr.utils.is_basic_terminal():
            console.line()
            console.rule(title="Summary", characters="=", style="[bold]rule.line")

            for renderable in self.summary_group.renderables:
                console.print(renderable)

            return []

        return super().__rich_console__(console, options)


def manifest_callback(ctx: Context, manifest_file: Path) -> Path:  # pragma: no cover
    """
    Add `manifest_file` value to shared Click context. Allows interspersed CLI options/arguments
    """
    ctx.params["manifest_file"] = manifest_file

    return manifest_file


@app.callback()
def bundle(  # pylint: disable=unused-argument, too-many-arguments
    manifest_file: Path = Argument(
        ...,
        callback=manifest_callback,
        dir_okay=False,
        exists=True,
        help="Path to manifest file",
        resolve_path=True,
        show_default=False,
    ),
    create_attestations: bool = Option(
        False,
        "-a",
        "--attest",
        help="Generate in-toto attestations",
        envvar="HOPPR_ATTESTATION",
        show_default=False,
    ),
    credentials_file: Optional[Path] = Option(
        None,
        "-c",
        "--credentials",
        help="Specify credentials config for services",
        envvar="HOPPR_CREDS_CONFIG",
        show_default=False,
    ),
    functionary_key_path: Optional[Path] = Option(
        None,
        "-fk",
        "--functionary-key",
        help="Path to key used to sign in-toto layout",
        envvar="HOPPR_FUNCTIONARY_KEY",
        show_default=False,
    ),
    functionary_key_password: Optional[str] = Option(
        None,
        "-fk-pass",
        "--functionary-key-password",
        confirmation_prompt=True,
        envvar="HOPPR_FUNCTIONARY_KEY_PASSWORD",
        help="Password for project owner key",
        show_default=False,
    ),
    functionary_key_prompt: bool = Option(
        False,
        "-p",
        "--prompt",
        help="Prompt user for functionary key's password",
        envvar="HOPPR_FUNCTIONARY_KEY_PROMPT",
    ),
    log_file: Optional[Path] = Option(
        None,
        "-l",
        "--log",
        help="File to which log will be written",
        envvar="HOPPR_LOG_FILE",
        show_default=False,
    ),
    previous_delivery: Optional[Path] = Option(
        None,
        "-pd",
        "--previous-delivery",
        help="Path to manifest or tar bundle representing a previous delivery",
        envvar="HOPPR_PREVIOUS_DELIVERY",
        show_default=False,
    ),
    transfer_file: Path = Option(
        "transfer.yml",
        "-t",
        "--transfer",
        help="Specify transfer config",
        envvar="HOPPR_TRANSFER_CONFIG",
        show_default=False,
    ),
    basic_term: bool = basic_term_option,
    experimental: bool = experimental_option,
    strict_repos: bool = strict_repos_option,
    verbose: bool = verbose_option,
):  # pragma: no cover
    """
    Run the stages specified in the transfer config
    file on the content specified in the manifest
    """
    processor.layout = HopprBundleLayout()
    processor.summary_panel = HopprBundleSummary()

    main.bundle(**locals())



"""
Base model for Hoppr config files
"""
from __future__ import annotations

import functools
import sys

from typing import Annotated, Any, Callable, ClassVar, Literal, MutableMapping, TypeVar

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from pydantic import BaseConfig, BaseModel, Extra, Field, create_model, validator
from pydantic_yaml import YamlModel

import hoppr.utils

AnyCycloneDXModel = TypeVar("AnyCycloneDXModel", bound="CycloneDXBaseModel")
UniqueIDMap = Annotated[MutableMapping[str, AnyCycloneDXModel], Field(default={})]

__all__ = []


class HopprMetadata(YamlModel):
    """
    Metadata data model
    """

    name: str
    version: str | int
    description: str


class HopprBaseModel(YamlModel):
    """
    Base Hoppr data model
    """

    class Config(BaseConfig):  # pylint: disable=too-few-public-methods
        """
        Config options for HopprBaseModel
        """

        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        extra = Extra.forbid
        use_enum_values = True

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        """
        Define to test equality or uniqueness between objects
        """
        return hash(repr(self))


class HopprBaseSchemaModel(HopprBaseModel):
    """
    Base Hoppr config file schema model
    """

    kind: Literal["Credentials", "Manifest", "Transfer"] = Field(..., description="Data model/schema kind")
    metadata: HopprMetadata | None = Field(None, description="Metadata for the file")
    schema_version: str = Field(..., alias="schemaVersion", title="Schema Version")

    @validator("kind", allow_reuse=True, pre=True)
    @classmethod
    def validate_kind(cls, kind: str) -> str:
        """
        Return supplied `kind` value with only first letter capitalized
        """
        return kind.capitalize()


class CycloneDXBaseModel(BaseModel):
    """
    Base CycloneDX data model
    """

    class Config(HopprBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config options for CycloneDXBaseModel"

    # Defining as ClassVar to allow dynamic model creation using custom root types
    deep_merge: ClassVar[bool] = False
    flatten: ClassVar[bool] = False
    observers: ClassVar[dict[object, Callable]] = {}
    unique_id_map: ClassVar[UniqueIDMap]

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other) if isinstance(other, type(self)) else False

    def __hash__(self) -> int:
        return hash(self.unique_id_callback())

    def __init__(self, **data):
        super().__init__(**data)

        unique_id = self.unique_id_callback()
        type(self).unique_id_map[unique_id] = self

    def _has_field(self, field_name: str) -> bool:
        return hasattr(self, field_name) and (getattr(self, field_name) or field_name in list(self.__fields_set__))

    @staticmethod
    def _is_list_of(obj_type: type, field: Any) -> bool:
        """
        Checks if the provided field is a list of the specified type
        """
        return isinstance(field, list) and all(isinstance(item, obj_type) for item in field)

    def _merge_field(self, target_field_name: str, source_field: Any) -> None:
        """
        Merges `source_field` into the field referenced by `target_field_name`
        """
        merged_field = getattr(self, target_field_name)
        field_type = type(merged_field)

        if not self._has_field(target_field_name):
            setattr(self, target_field_name, source_field)
            return

        if isinstance(field_type, type(BaseModel)):
            merged_field = CycloneDXBaseModel.create(model=merged_field)
            source_field = CycloneDXBaseModel.create(model=source_field)
            merged_field.merge(source_field)
        elif self._is_list_of(BaseModel, merged_field):
            merged_field = hoppr.utils.dedup_list([CycloneDXBaseModel.create(model=_field) for _field in merged_field])
            source_field = [CycloneDXBaseModel.create(model=_field) for _field in source_field or []]
            self._merge_field_items(merged_field, source_field)

        setattr(self, target_field_name, merged_field)

    @staticmethod
    def _merge_field_items(target_field: list[CycloneDXBaseModel], source_field: list[CycloneDXBaseModel]) -> None:
        """
        Merges the items from `source_field` into `target_field`
        """
        for source_item in source_field:
            if source_item in target_field:
                merged_item = target_field[target_field.index(source_item)]
                merged_item.merge(source_item)
            else:
                target_field.append(source_item)

    def unique_id_callback(self) -> str:
        """
        Default callback method to get a model object's unique ID
        """
        try:
            callback = {
                "Advisory": lambda obj: obj.url,
                "Annotations": lambda obj: obj.bom_ref or obj.text,
                "Command": lambda obj: obj.executed or repr(self),
                "Commit": lambda obj: obj.uid,
                "Component": lambda obj: obj.bom_ref,
                "ComponentData": lambda obj: obj.bom_ref or repr(self),
                "Copyright": lambda obj: obj.text,
                "Dependency": lambda obj: obj.ref,
                "ExternalReference": lambda obj: f"{obj.type}-{obj.url}",
                "Formula": lambda obj: obj.bom_ref or repr(self),
                "License": lambda obj: obj.id or obj.name or repr(self),
                "LicenseChoice": lambda obj: obj.license or obj.expression or repr(self),
                "Note": lambda obj: obj.text,
                "Occurrence": lambda obj: obj.bom_ref or repr(self),
                "Reference": lambda obj: obj.id,
                "Sbom": lambda obj: obj.serialNumber,
                "Service": lambda obj: obj.bom_ref or f"{obj.name}-{obj.version}",
                "Signer": lambda obj: obj.value,
                "Task": lambda obj: obj.bom_ref or obj.uid,
                "Vulnerability": lambda obj: obj.id or repr(self),
                "Workflow": lambda obj: obj.bom_ref or obj.uid,
                "Workspace": lambda obj: obj.bom_ref or obj.uid,
            }[type(self).__name__]

            return callback(self)
        except KeyError:
            return repr(self)

    @classmethod
    def create(cls: type[AnyCycloneDXModel], model: BaseModel) -> AnyCycloneDXModel:
        """
        Update a BaseModel object with CycloneDXBaseModel attributes and methods

        Args:
            model (BaseModel): The BaseModel object to update

        Returns:
            AnyCycloneDXModel: The updated BaseModel object
        """
        model_cls = cls.make_model(name=type(model).__name__)
        return model_cls(**model.dict(by_alias=True, exclude_none=True, exclude_unset=True))  # pyright: ignore

    @classmethod
    @functools.cache
    def make_model(cls: type[AnyCycloneDXModel], name: str) -> type[AnyCycloneDXModel]:
        """
        Dynamically create a model class suitable for merging

        Args:
            name (str): Name of the existing model

        Returns:
            type[AnyCycloneDXModel]: The generated model class
        """
        # Return explicitly defined models directly
        if name in {"Component", "ExternalReference", "Hash", "Property"}:
            return sys.modules["hoppr.models.sbom"].__dict__[name]

        if name.startswith("License"):
            return sys.modules["hoppr.models.licenses"].__dict__[name]

        if name.startswith("Affect"):
            return sys.modules["hoppr.models.affect"].__dict__[name]

        model_cls = cdx.__dict__[name]

        merge_model = create_model(model_cls.__name__, __base__=(cls, model_cls), __module__=__name__)

        # Set model's pydantic `Config` class and `__hash__` method
        setattr(merge_model, "__config__", cls.Config)

        # Add `unique_id_map` class attribute for caching model objects
        merge_model.__class_vars__.add("unique_id_map")
        merge_model.__annotations__["unique_id_map"] = "ClassVar[UniqueIDMap]"
        setattr(merge_model, "unique_id_map", {})

        # Add updated model to current module and make importable from other modules
        setattr(sys.modules[__name__], merge_model.__name__, merge_model)
        __all__.append(merge_model.__name__)
        merge_model.update_forward_refs()

        return merge_model

    @classmethod
    def find(cls: type[AnyCycloneDXModel], unique_id: str) -> AnyCycloneDXModel | None:
        """
        Look up model object by its unique ID string

        Args:
            unique_id (str): Unique ID string to look up

        Returns:
            AnyCycloneDXModel | None: Model object if found, otherwise None
        """
        return cls.unique_id_map.get(unique_id)

    def merge(self: AnyCycloneDXModel, other: AnyCycloneDXModel) -> None:
        """
        Merge model instance of same type into self

        Args:
            other (AnyCycloneDXModel): Model object to merge
        """
        if (self_type := type(self).__name__) != (other_type := type(other).__name__):
            raise TypeError(f"Type '{other_type}' cannot be merged into '{self_type}'")

        self.notify(data=f"  Merging '{type(self).__qualname__}' attributes...")

        for field_name in self.__fields__:
            self.notify(data=f"    Merging field '{type(self).__qualname__}.{field_name}'...")

            if (source_field := getattr(other, field_name, None)) is None:
                continue

            self._merge_field(field_name, source_field)

    def notify(self, data: str) -> None:
        """
        Call the callback function for all registered subscribers
        """
        for callback in self.observers.values():
            callback(data)

    def subscribe(self, observer: object, callback: Callable) -> None:
        """
        Register an observer
        """
        self.observers[observer] = callback

    def unsubscribe(self, observer: object) -> None:
        """
        Unregister an observer
        """
        self.observers.pop(observer, None)



"""
Collector plugin for Cargo packages
"""
from __future__ import annotations
from pydantic import SecretStr

import requests

from requests import HTTPError
from requests.auth import HTTPBasicAuth

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectCargoPlugin(SerialCollectorPlugin):
    """
    Collector plugin for Cargo packages
    """

    supported_purl_types = ["cargo"]
    products: list[str] = ["cargo/*"]
    system_repositories: list[str] = ["https://crates.io/api/v1/crates"]

    def get_version(self):
        return __version__

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        # pylint: disable=duplicate-code
        purl = hoppr.utils.get_package_url(comp.purl)
        cargo_url = RepositoryUrl(url=repo_url) / purl.name / (purl.version or "") / "download"

        self.get_logger().info(msg=f"Copying Cargo package from {cargo_url}", indent_level=2)

        authentication = None
        if creds is not None and isinstance(creds.password, SecretStr):
            authentication = HTTPBasicAuth(creds.username, creds.password.get_secret_value())

        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        package_out_path = target_dir / f"{purl.name}-{purl.version}.crate"

        response = requests.get(
            str(cargo_url),
            auth=authentication,
            timeout=300,
            proxies=self._get_proxies_for(repo_url),
            stream=True,
        )

        try:
            response.raise_for_status()
        except HTTPError:
            msg = f"Failed to download Cargo package {comp.purl}, status_code={response.status_code}"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.fail(message=msg)

        with package_out_path.open(mode="ab") as out_file:
            for chunk in response.iter_content(chunk_size=None):
                out_file.write(chunk)

        self.set_collection_params(comp, repo_url, target_dir)
        return Result.success(return_obj=comp)



"""Functions to faciliate interaction with non-image OCI Artifacts"""
import re

from pathlib import Path
from typing import Dict, Tuple

from oras.client import OrasClient
from oras.container import Container
from oras.provider import Registry

from pydantic import SecretStr

from hoppr import utils
from hoppr.exceptions import HopprLoadDataError
from hoppr.models.credentials import Credentials

# Regex per: https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
SEM_VER_CORE_REGEX = r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)$"
SEM_VER_FULL_REGEX = (
    r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
    r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)


def pull_artifact(artifact: str, allow_version_discovery=False, version_regex=SEM_VER_CORE_REGEX) -> Dict:
    """
    Pulls OCI Artifact from registry into memory and returns content

    If no version (tag) is provided on URL, hoppr can attempt to disover highest version if allow_version_discovery
    is enabled. A regex for match versions can be provided to handle less standard version formats.
    Default will look for SemVer2 core
    """
    pulled = _pull_artifact(artifact, allow_version_discovery=allow_version_discovery, version_regex=version_regex)

    # We should only get one. If we get more, just take the first
    loaded_content = utils.load_file(pulled)
    if not isinstance(loaded_content, dict):
        raise HopprLoadDataError("Pulled artifact data was not loaded as dictionary")

    return loaded_content


def pull_artifact_to_disk(
    artifact: str,
    target_dir: str,
    allow_version_discovery=False,
    version_regex=SEM_VER_CORE_REGEX,
):
    """Pulls OCI Artifact from registry to specified target_dir.

    If no version (tag) is provided on URL, hoppr can attempt to disover highest version if allow_version_discovery
    is enabled. A regex for match versions can be provided to handle less standard version formats.
    Default will look for SemVer2 core
    """
    return _pull_artifact(
        artifact,
        target_dir,
        allow_version_discovery=allow_version_discovery,
        version_regex=version_regex,
    )


def _create_client_info(artifact: str) -> Tuple[Container, OrasClient]:
    """Utility function for creating ORAS client and intermediate objects"""
    # Use ORAS' Container object to faciliate parsing registry host
    container = Container(artifact)
    registry = Registry(hostname=container.registry)

    creds = Credentials.find(container.registry)
    if creds is not None and isinstance(creds.password, SecretStr):
        registry.set_basic_auth(creds.username, creds.password.get_secret_value())
    client = OrasClient(registry=registry)

    return container, client


def _pull_artifact(
    artifact: str,
    outdir=None,
    allow_version_discovery=False,
    version_regex=SEM_VER_CORE_REGEX,
) -> Path:
    """Looks up and pulls artifact to disk from OCI registry"""
    container, client = _create_client_info(artifact)

    full_artifact = artifact
    # Only discover tag if it hasn't been explicitly defined.
    # We let the 'Container' do the parsing, so we'll just compare
    if not artifact.endswith(container.tag):
        if not allow_version_discovery:
            raise HopprLoadDataError(
                f"No version provided and version discovery is disabled. Cannot pull artifact {artifact}"
            )
        version_tag = _discover_highest_version_tag(artifact, version_regex=version_regex)
        full_artifact += f":{version_tag}"

    files = client.pull(target=full_artifact, outdir=outdir)
    return Path(files[0])


def _discover_highest_version_tag(artifact: str, version_regex=SEM_VER_CORE_REGEX):
    """Attempts to discover the highest versioned artifact based on provided
    regex. Default will use full SemVer"""
    container, client = _create_client_info(artifact)

    tags: list[str] = client.get_tags(str(container))

    # Only find those that match semver
    # Future enhancement can allow user to define regex to use in the artifact definition
    tags = sorted(filter(lambda x: re.match(version_regex, x), tags["tags"]))  # type: ignore[arg-type,call-overload]

    return sorted(tags)[-1]



"""
Enumeration to indicate how a plugin may change a BOM
"""
from __future__ import annotations

import re

from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator

from hoppr_cyclonedx_models.cyclonedx_1_5 import Component
from hoppr_cyclonedx_models.cyclonedx_1_5 import CyclonedxSoftwareBillOfMaterialsStandard as Bom
from pydantic import AnyUrl, Field, FileUrl, HttpUrl, SecretStr, root_validator
from pydantic_yaml import YamlIntEnum, YamlStrEnum

from hoppr.models.base import HopprBaseModel

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny


class LocalFile(HopprBaseModel):
    """
    LocalFile data model
    """

    local: Path

    def __str__(self) -> str:
        return self.local.resolve().as_uri()


class OciFile(HopprBaseModel):
    """
    OciFile data model
    """

    oci: AnyUrl | str

    def __str__(self) -> str:
        return str(self.oci)


class UrlFile(HopprBaseModel):
    """
    UrlFile data model
    """

    url: HttpUrl | FileUrl | AnyUrl | str

    def __str__(self) -> str:
        return str(self.url)


class BomAccess(YamlIntEnum):
    """
    Enumeration to indicate how a plugin may change a BOM
    """

    # pylint: disable=duplicate-code
    NO_ACCESS = 0
    COMPONENT_ACCESS = 1
    FULL_ACCESS = 2

    def has_access_to(self, obj: Any) -> bool:
        """
        Determine whether the specified value allows updates to an object
        """
        match self:
            case BomAccess.NO_ACCESS:
                return obj is None
            case BomAccess.COMPONENT_ACCESS:
                return isinstance(obj, Component)
            case BomAccess.FULL_ACCESS:
                return isinstance(obj, Bom)


class PurlType(YamlStrEnum):
    """
    Enumeration of supported purl types
    """

    # pylint: disable=duplicate-code
    CARGO = "Cargo"
    DEB = "Deb"
    DOCKER = "Docker"
    GEM = "Gem"
    GENERIC = "Generic"
    GIT = "Git"
    GITHUB = "GitHub"
    GITLAB = "GitLab"
    GOLANG = "Golang"
    HELM = "Helm"
    MAVEN = "Maven"
    NPM = "NPM"
    NUGET = "NuGet"
    OCI = "OCI"
    PYPI = "PyPI"
    RAW = "Raw"
    REPO = "Repo"
    RPM = "RPM"

    def __str__(self) -> str:
        return self.value.lower()

    @classmethod
    def _missing_(cls, value: object):
        return next((member for member in cls if member.name == str(value).upper()), None)

    @classmethod
    def values(cls) -> Iterator[str]:
        """
        Iterate over enum member strings.

        Yields:
            Iterator[str]: Iterator of enum member strings.
        """
        yield from [str(member) for member in cls]


class RepositoryUrl(HopprBaseModel):
    """
    RepositoryUrl data model
    """

    class Config(HopprBaseModel.Config):  # pylint: disable=too-few-public-methods
        "RepositoryUrl model config"
        allow_mutation = False
        validate_assignment = True

    url: str = Field(...)
    scheme: str | None = Field(default=None)
    username: str | None = Field(default=None)
    password: SecretStr | str | None = Field(default=None)
    hostname: str | None = Field(default=None)
    port: int | None = Field(default=None)
    path: str | None = Field(default=None)
    query: str | None = Field(default=None)
    fragment: str | None = Field(default=None)
    netloc: str | None = Field(default=None, exclude=True)

    @root_validator(pre=True)
    @classmethod
    def validate_model(cls, values: DictStrAny):
        """
        Validate RepositoryUrl model
        """
        if not (url := values.get("url")):
            raise ValueError("Input parameter `url` must be a non-empty string")

        if url.endswith(":"):
            url = f"{url}//"

        if not (
            match := re.search(
                pattern=(
                    r"^((?P<scheme>[^:/?#]+):(?=//))?(//)?((("
                    r"?P<username>[^:]+)(?::("
                    r"?P<password>[^@]+)?)?@)?("
                    r"?P<hostname>[^@/?#:]*)(?::("
                    r"?P<port>\d+)?)?)?("
                    r"?P<path>[^?#]*)(\?("
                    r"?P<query>[^#]*))?(#("
                    r"?P<fragment>.*))?"
                ),
                string=url,
            )
        ):
            raise ValueError(f"Not a valid URL: {url}")

        values.update({name: value for name, value in match.groupdict().items() if value is not None})

        if values.get("port"):
            values["port"] = int(values["port"])

        if isinstance(password := values.get("password"), str):
            password = SecretStr(password)

        auth = ":".join(filter(None, [values.get("username"), values.get("password")]))
        host_port = ":".join(filter(None, [values.get("hostname"), str(values.get("port") or "")]))
        values["netloc"] = "@".join(filter(None, [auth, host_port]))

        values["url"] = f"{values['scheme']}://" if values.get("scheme") else ""
        values["url"] = f"{values['url']}{values['netloc']}"
        values["url"] = "/".join(filter(None, [values["url"], values["path"].strip("/") if values.get("path") else ""]))
        values["url"] = "?".join(filter(None, [values["url"], values.get("query")]))
        values["url"] = "#".join(filter(None, [values["url"], values.get("fragment")]))

        return values

    def __repr__(self) -> str:
        props = ", ".join(f"{name}={repr(value)}" for name, value in dict(self).items())
        return f"RepositoryUrl({props})"

    def __str__(self) -> str:
        return self.url

    def __truediv__(self, other: str) -> RepositoryUrl:
        return self.join(other)

    def __itruediv__(self, other: str) -> RepositoryUrl:
        # Convert all instances of two or more slashes to single slash
        other = "/".join(filter(None, other.split("/")))
        return self.join(other)

    def join(self, other: str) -> RepositoryUrl:  # pylint: disable=missing-docstring
        # Convert two or more slashes to single slash (ignores `://`)
        joined = "/".join([re.sub(pattern="(?<!:)//+", repl="/", string=self.url), other.strip("/")])
        return type(self)(url=joined)



"""
Hoppr utility functions
"""
from __future__ import annotations

import functools
import importlib
import inspect
import os
import pyclbr
import sys

from importlib.metadata import entry_points
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Iterable, Type
from urllib.parse import unquote

import psutil

from packageurl import PackageURL
from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError as YAMLParserError
from ruamel.yaml.scanner import ScannerError as YAMLScannerError

from hoppr.exceptions import HopprError, HopprLoadDataError, HopprPluginError

if TYPE_CHECKING:
    from hoppr.base_plugins.hoppr import HopprPlugin
    from hoppr.models import HopprContext


def _class_in_module(plugin_module: ModuleType) -> Type[HopprPlugin]:
    """
    Find the appropriate class in given module

    Args:
        plugin_module (ModuleType): Previously loaded module to search for specified plugin class

    Raises:
        HopprPluginError: No class definition or more than one class definition found in module

    Returns:
        Type[HopprPlugin]: Class type of plugin
    """
    # Prevent circular import
    base_cls = getattr(importlib.import_module(name="hoppr.base_plugins.hoppr"), "HopprPlugin")

    # Get classes defined in plugin_module
    module_members = pyclbr.readmodule(module=plugin_module.__name__)

    def is_hoppr_plugin(obj: ModuleType):
        return inspect.isclass(obj) and issubclass(obj, base_cls) and obj.__name__ in module_members

    module_classes = dict(inspect.getmembers(object=plugin_module, predicate=is_hoppr_plugin))

    match list(module_classes.keys()):
        case []:
            raise HopprPluginError(f"No class definition found in '{plugin_module.__name__}'.")
        case [multiple_cls, *extra] if extra:
            cls_names = ", ".join(f"'{cls_name}'" for cls_name in [multiple_cls, *extra])
            raise HopprPluginError(f"Multiple candidate classes defined in '{plugin_module.__name__}': {cls_names}")

    return list(module_classes.values())[0]


def plugin_class(plugin_name: str) -> Type[HopprPlugin]:
    """
    Get plugin class type from module path or class name

    If the plugin name references a module, it is assumed the module defines
    exactly one subclass of `HopprPlugin`, which will be instantiated using
    a default constructor (i.e., one with no parameters).

    Args:
        plugin_name (str): Name of the plugin module or class defined in transfer file

    Raises:
        ModuleNotFoundError: Import of specified module failed
        HopprPluginError: No class definition or more than one class definition found in module

    Returns:
        Type[HopprPlugin]: Class type of plugin
    """
    discovered_plugins = entry_points(group="hoppr.plugin")

    try:
        if found_plugin := [
            *discovered_plugins.select(name=plugin_name),
            *discovered_plugins.select(module=plugin_name),
            *discovered_plugins.select(attr=plugin_name),
        ]:
            match found_plugin:
                case [found_one]:
                    loaded = found_one.load()

                    # Return directly if provided plugin name was loaded as a plugin class
                    # Otherwise, search the loaded plugin module for appropriate class
                    return loaded if inspect.isclass(loaded) else _class_in_module(loaded)
                case [found_multiple, *extra]:
                    matches = ", ".join(f"'{found.name}'" for found in [found_multiple, *extra])
                    raise HopprPluginError(f"Multiple entry points matched '{plugin_name}': {matches}")

        # Provided plugin name not found in entry points
        # Attempt to load using fully qualified module path
        loaded = importlib.import_module(name=plugin_name)
        return _class_in_module(loaded)
    except ModuleNotFoundError as mnfe:
        raise ModuleNotFoundError(f"Unable to locate plug-in '{plugin_name}': {mnfe}") from mnfe
    except HopprPluginError as ex:
        raise ex


def plugin_instance(plugin_name: str, context: HopprContext, config: Any = None):
    """
    Create an instance of an object defined by a plugin name

    Assumes the specified plugin will define exactly one concrete class, which
    will be instaniated using a default constructor (i.e., one with no parameters).
    """
    plugin_cls = plugin_class(plugin_name)

    instance = plugin_cls(context=context, config=dict(config or {}))

    # If `instance.supported_purl_types` is empty, all PURL types are supported
    if not instance.supported_purl_types:
        purl_types = getattr(importlib.import_module("hoppr.models.types"), "PurlType")
        instance.supported_purl_types = [str(purl_type) for purl_type in purl_types]

    return instance


def load_string(contents: str) -> dict | list | None:
    """
    Return a YAML or JSON formatted string as a dict
    """
    if not contents.strip():
        raise HopprLoadDataError("Empty string cannot be parsed.")

    # Replace tab characters with spaces to prevent parsing errors
    contents = contents.replace("\t", "  ")
    loaded_contents: dict | list | None = None

    try:
        # Applicable to both YAML and JSON formats since valid JSON data is also valid YAML
        yaml = YAML(typ="safe", pure=True)
        loaded_contents = yaml.load(contents)

        # yaml.safe_load will sometimes return a single string rather than the required structure
        if isinstance(loaded_contents, str):
            raise HopprLoadDataError("Expected dictionary or list, but contents were loaded and returned as string")
    except (YAMLParserError, YAMLScannerError) as ex:
        raise HopprLoadDataError("Unable to recognize data as either json or yaml") from ex
    except HopprLoadDataError as ex:
        raise ex

    return loaded_contents


def load_file(input_file_path: Path) -> dict | list | None:
    """
    Load file content (either JSON or YAML) into a dict
    """

    if not input_file_path.is_file():
        raise HopprLoadDataError(f"{input_file_path} is not a file, cannot be opened.")

    with input_file_path.open(mode="r", encoding="utf-8") as input_file:
        content: str = input_file.read()
        if not content.strip():
            raise HopprLoadDataError(f"File {input_file_path} is empty.")

    return load_string(content)


def dedup_list(list_in: Iterable[Any]) -> list[Any]:
    """
    De-duplicate a list
    """
    return list(dict.fromkeys(list_in or []))


def obscure_passwords(command_list: list[str], sensitive_args: list[str] | None = None) -> str:
    """
    Returns an input string with any specified passwords hidden
    """

    password_list: list[str] = sensitive_args if sensitive_args is not None else []
    obscured_command_list: list[str] = []

    for arg in command_list:
        # Quote arguments that contain spaces
        if " " in arg:
            arg = f'"{arg}"'

        # Replace password string(s) in argument
        for password in password_list:
            if password is not None and len(password) > 0:
                arg = arg.replace(password, "[masked]")

        obscured_command_list.append(arg)

    return " ".join(obscured_command_list)


def remove_empty(directory: Path) -> set[Path]:
    """
    Removes empty folders given the directory including parent folders
    """
    deleted: set[Path] = set()

    if not directory.exists():
        raise FileNotFoundError()

    for subdir in directory.iterdir():
        if subdir.is_dir():
            deleted.update(remove_empty(subdir))

    if directory.is_dir() and not any(directory.iterdir()):
        directory.rmdir()
        deleted.add(directory)

    return deleted


@functools.cache
def get_package_url(purl_string: str) -> PackageURL:
    """
    Get the PackageURL for a given purl_string and store it in a cache for improved performance.

    Args:
        purl_string (str): The string representation of a Package URL

    Returns:
        PackageURL: The object representation of a PackageURL

    """
    return PackageURL.from_string(unquote(purl_string))


def get_partition(directory: os.PathLike) -> psutil._common.sdiskpart:  # pylint: disable=protected-access
    """
    Returns partition information for a file location
    """

    mount = Path(directory).expanduser().resolve()
    while not mount.is_mount():
        mount = mount.parent

    try:
        return next((part for part in psutil.disk_partitions(all=True) if part.mountpoint == str(mount)))
    except StopIteration as ex:
        raise HopprError(f"Unable to identify partition for path '{directory}'") from ex


def is_cicd() -> bool:
    """
    Check if running in CI/CD context
    """
    return any(
        os.getenv(env_var, None)
        for env_var in [
            "bamboo.buildKey",
            "BUILD_ID",
            "BUILD_NUMBER",
            "BUILDKITE",
            "CI",
            "CIRCLECI",
            "CIRRUS_CI",
            "CODEBUILD_BUILD_ID",
            "CONTINUOUS_INTEGRATION",
            "GITHUB_ACTIONS",
            "GITLAB_CI",
            "HEROKU_TEST_RUN_ID",
            "HUDSON_URL",
            "JENKINS_URL",
            "TEAMCITY_VERSION",
            "TF_BUILD",
            "TRAVIS",
        ]
    )


@functools.cache
def is_basic_terminal() -> bool:
    """
    Check if current terminal supports ANSI multi-byte characters

    This function is decorated with `functools.cache` so that it always returns the same value after the initial
    invocation. This is to prevent negating the return value once the Rich live display is started, i.e.
    `sys.stdout.isatty()` will then return False instead of True
    """
    return any(
        [
            is_cicd(),
            not sys.stdout.isatty(),
            os.getenv("HOPPR_BASIC_TERM"),
            os.getenv("TERM_PROGRAM") == "Apple_Terminal",
        ]
    )



"""
Common options for typer commands
"""
import os

from typer import CallbackParam, Context, Option


def set_global_option(ctx: Context, param: CallbackParam, value: bool) -> bool:
    """
    Add option value to shared Click context. Allows interspersed CLI options/arguments

    Args:
        ctx (Context): Click context to update
        param (CallbackParam): Typer metadata for the parameter
        value (bool): The parameter value

    Returns:
        bool: The flag that was passed, unmodified
    """
    if param.name:
        ctx.params[param.name] = value

        if param.envvar and value:
            os.environ[str(param.envvar)] = "1"

    return value


basic_term_option: bool = Option(
    False,
    "-b",
    "--basic-term",
    callback=set_global_option,
    help="Use simplified output for non-TTY or legacy terminal emulators",
    is_eager=True,
    envvar="HOPPR_BASIC_TERM",
    rich_help_panel="Global options",
    show_default=False,
)

experimental_option: bool = Option(
    False,
    "-x",
    "--experimental",
    callback=set_global_option,
    help="Enable experimental features",
    is_eager=True,
    envvar="HOPPR_EXPERIMENTAL",
    rich_help_panel="Global options",
    show_default=False,
)

strict_repos_option: bool = Option(
    True,
    "--strict/--no-strict",
    callback=set_global_option,
    help="Utilize only manifest repositories for package collection",
    is_eager=True,
    envvar="HOPPR_STRICT_REPOS",
    rich_help_panel="Global options",
    show_default=False,
)

verbose_option: bool = Option(
    False,
    "-v",
    "--debug",
    "--verbose",
    callback=set_global_option,
    help="Enable debug output",
    is_eager=True,
    rich_help_panel="Global options",
    show_default=False,
)



"""
`merge` subcommand for `hoptctl`
"""
from __future__ import annotations

import time

from pathlib import Path
from typing import TYPE_CHECKING

import typer

from pydantic import ValidationError
from rich.console import Group
from rich.live import Live
from rich.progress import Progress
from rich.table import Table
from typer import BadParameter, Option, Typer

import hoppr.utils

from hoppr.cli.layout import HopprBasePanel, HopprJobsPanel, HopprLayout, HopprSpinnerColumn, console
from hoppr.cli.options import basic_term_option, verbose_option
from hoppr.models.base import CycloneDXBaseModel
from hoppr.models.manifest import Manifest
from hoppr.models.sbom import Sbom

if TYPE_CHECKING:
    from rich.console import Console, ConsoleOptions, RenderResult

app = Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Merge all properties of two or more SBOM files",
    invoke_without_command=True,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="markdown",
    subcommand_metavar="",
)


class HopprMergeJobsPanel(HopprJobsPanel):
    """
    Customized Rich Progress bar Panel
    """

    progress_bar = Progress(
        "{task.description}",
        HopprSpinnerColumn(),
        expand=True,
    )


class HopprMergeLayout(HopprLayout):
    """
    Layout of the `hopctl merge` console application
    """

    name: str = "root"
    jobs_panel = HopprMergeJobsPanel()


class HopprMergeSummary(HopprBasePanel):  # pylint: disable=too-few-public-methods
    """
    `hopctl merge` results summary
    """

    stage_results_map: dict[str, Table] = {}

    def __init__(self, *args, title: str | None = "Summary", **kwargs) -> None:
        self.total_components = 0
        self.merged_components = 0

        self.summary_table = Table(
            "File Name",
            "# Components",
            "Elapsed Time",
            box=None,
            pad_edge=False,
            show_edge=False,
        )

        self.before_after_table = Table.grid(padding=(0, 1))

        self.summary_group = Group(self.summary_table, self.before_after_table)

        super().__init__(self.summary_group, title, *args, **kwargs)

    def __rich_console__(self, console_: Console, options: ConsoleOptions) -> RenderResult:
        # pylint: disable=arguments-renamed
        self.before_after_table.add_row()
        self.before_after_table.add_row("Total components before merge:", str(self.total_components))
        self.before_after_table.add_row("Total components after merge:", str(self.merged_components))

        if hoppr.utils.is_basic_terminal():
            console_.line()
            console_.rule(title="Summary", characters="=", style="[bold]rule.line")
            console_.print(self.summary_table)
            console_.print(self.before_after_table)
            return []

        return super().__rich_console__(console_, options)


def manifest_callback(manifest_file: Path | None) -> Path | None:
    """
    Extract and return SBOM refs from manifest files
    """
    if manifest_file is None:
        return None

    # Load manifest file to populate `Sbom.loaded_sboms`
    Manifest.load(manifest_file)

    return manifest_file


def output_file_callback(output_file: Path | None) -> Path:
    """
    Auto-generate an output file name if not provided
    """
    return output_file or Path.cwd() / f"hopctl-merge-{time.strftime('%Y%m%d-%H%M%S')}.json"


def sbom_callback(sbom_sources: list[str | Path]) -> list[str | Path]:
    """
    Load SBOM input files
    """
    # Combine SBOM files from all CLI input file arguments
    _load_sbom_files(sbom_sources)

    return sbom_sources


def sbom_dir_callback(sbom_dirs: list[Path]) -> list[Path]:
    """
    Load SBOM files from input directories
    """
    for sbom_dir in sbom_dirs:
        sbom_files = sbom_dir.glob("*.json")

        _load_sbom_files(list(sbom_files))

    return sbom_dirs


def _load_sbom_files(sbom_files: list[str | Path]) -> None:
    """
    Load SBOM input files
    """
    for sbom_file in sbom_files:
        try:
            Sbom.load(sbom_file)
        except ValidationError as ex:
            raise BadParameter(f"'{sbom_file}' is not a valid SBOM file") from ex


def _merge_sboms() -> Sbom:
    """
    Merge SBOMs into single object
    """
    merged_sbom = Sbom()
    layout = HopprMergeLayout()
    live_display = Live(layout, console=console, refresh_per_second=10)
    merged_sbom.subscribe(observer=layout, callback=layout.print)
    summary_panel = HopprMergeSummary()

    for sbom_ref, sbom in Sbom.loaded_sboms.items():
        summary_panel.total_components += len(sbom.components)
        layout.add_job(description=Path(str(sbom_ref)).name, sbom=sbom)

    if not hoppr.utils.is_basic_terminal():
        live_display.start(refresh=True)

    for task in layout.jobs_panel.progress_bar.tasks:
        layout.print(f"Merging {task.description}...")
        layout.start_job(task.description)

        merged_sbom.merge(task.fields["sbom"])

        layout.stop_job(task.description)
        layout.advance_job(task.description)

        summary_panel.summary_table.add_row(
            task.description, str(len(task.fields["sbom"].components)), f"{task.elapsed:.3f}"
        )

    layout.print("Writing merged SBOM...")

    layout.overall_progress.progress_bar.stop_task(layout.progress_task.id)

    summary_panel.merged_components = len(merged_sbom.components)

    if not hoppr.utils.is_basic_terminal():
        live_display.stop()

    console.print("\n", summary_panel)

    merged_sbom.unsubscribe(observer=layout)
    return merged_sbom


@app.callback()
def merge(  # pylint: disable=too-many-arguments, unused-argument
    manifest_file: Path = Option(
        None,
        "-m",
        "--manifest",
        callback=manifest_callback,
        dir_okay=False,
        exists=True,
        help="Manifest file containing SBOMs to merge",
        resolve_path=True,
        show_default=False,
    ),
    sbom_files: list[Path] = Option(
        [],
        "-s",
        "--sbom",
        callback=sbom_callback,
        dir_okay=False,
        exists=True,
        help="SBOM file to merge (can be specified multiple times)",
        resolve_path=True,
        show_default=False,
    ),
    sbom_dirs: list[Path] = Option(
        [],
        "-d",
        "--sbom-dir",
        callback=sbom_dir_callback,
        exists=True,
        file_okay=False,
        help="Directory containing SBOM files to merge (can be specified multiple times)",
        resolve_path=True,
        show_default=False,
    ),
    sbom_urls: list[str] = Option(
        [],
        "-u",
        "--sbom-url",
        callback=sbom_callback,
        help="URL of SBOM to merge (can be specified multiple times)",
        metavar="URL",
        show_default=False,
    ),
    output_file: Path = Option(
        None,
        "-o",
        "--output-file",
        callback=output_file_callback,
        dir_okay=False,
        exists=False,
        help=f"Path to output file {typer.style(text='[default: hopctl-merge-YYMMDD-HHMMSS.json]', dim=True)}",
        resolve_path=True,
        show_default=False,
    ),
    deep_merge: bool = Option(
        False,
        "--deep-merge",
        help="Resolve and expand `externalReferences` in-place",
        show_default=False,
    ),
    flatten: bool = Option(
        False,
        "--flatten",
        help="Flatten nested `components` into single unified list",
        show_default=False,
    ),
    basic_term: bool = basic_term_option,
    verbose: bool = verbose_option,
):
    """
    Merge SBOM files
    """
    if not any([manifest_file, *sbom_dirs, *sbom_files, *sbom_urls]) or len(Sbom.loaded_sboms) < 2:
        raise BadParameter(
            "A minimum of two SBOM files must be provided via the "
            "--sbom, --sbom-dir, --sbom-url, and --manifest arguments."
        )

    CycloneDXBaseModel.deep_merge = deep_merge
    CycloneDXBaseModel.flatten = flatten

    merged_sbom = _merge_sboms()

    output_file.write_text(
        data=merged_sbom.json(exclude_none=True, exclude_unset=True, by_alias=True, indent=2),
        encoding="utf-8",
    )



"""
Class to store results of processes
"""
from __future__ import annotations

from enum import IntEnum
from typing import Any

import requests

from hoppr.exceptions import HopprError


class ResultStatus(IntEnum):
    """
    Enumeration of possible result states
    """

    EXCLUDED = 0
    SUCCESS = 1
    RETRY = 2
    FAIL = 3
    SKIP = 4


class Result:
    """
    Class to store results of processes
    """

    def __init__(self, status: ResultStatus, message: str = "", return_obj: Any = None):
        self.status = status
        self.message = message
        self.return_obj = return_obj

    def __str__(self):
        result_msg = f"{self.status.name}"
        if self.message != "":
            result_msg += f", msg: {self.message}"

        return result_msg.rstrip()

    @staticmethod
    def success(message="", return_obj: Any = None) -> Result:
        """
        Convenience method for generating success messages
        """
        return Result(ResultStatus.SUCCESS, message, return_obj)

    @staticmethod
    def retry(message="", return_obj: Any = None) -> Result:
        """
        Convenience method for generating retry messages
        """
        return Result(ResultStatus.RETRY, message, return_obj)

    @staticmethod
    def fail(message="", return_obj: Any = None) -> Result:
        """
        Convenience method for generating failure messages
        """
        return Result(ResultStatus.FAIL, message, return_obj)

    @staticmethod
    def excluded(message="", return_obj: Any = None) -> Result:
        """
        Convenience method for generating excluded messages
        """
        return Result(ResultStatus.EXCLUDED, message, return_obj)

    @staticmethod
    def skip(message="", return_obj: Any = None) -> Result:
        """
        Convenience method for generating skip messages
        """
        return Result(ResultStatus.SKIP, message, return_obj)

    def is_success(self) -> bool:
        """
        Convenience method for testing for success messages
        """
        return self.status == ResultStatus.SUCCESS

    def is_retry(self) -> bool:
        """
        Convenience method for testing for retry messages
        """
        return self.status == ResultStatus.RETRY

    def is_fail(self) -> bool:
        """
        Convenience method for testing for failure messages
        """
        return self.status == ResultStatus.FAIL

    def is_excluded(self) -> bool:
        """
        Convenience method for testing for excluded messages
        """
        return self.status == ResultStatus.EXCLUDED

    def is_skip(self) -> bool:
        """
        Convenience method for testing for skip messages
        """
        return self.status == ResultStatus.SKIP

    def merge(self, other: Result):
        """
        Logically combine two Result objects
        """
        if other.is_skip():
            return

        self.status = max(self.status, other.status)
        if self.message == "":
            self.message = other.message
        else:
            self.message += "\n" + other.message

        if self.return_obj is None:
            self.return_obj = other.return_obj
        elif other.return_obj is not None:
            raise HopprError("Unable to merge two results with return objects")

        return

    @staticmethod
    def from_http_response(response: requests.Response, return_obj: Any = None):
        """
        Build a Result object from an HTTP request response
        """

        match response.status_code:
            case response_code if 200 <= response_code <= 299:
                return Result.success(
                    f"HTTP Status Code: {response.status_code}",
                    return_obj=return_obj,
                )
            case response_code if response_code >= 500:
                status = ResultStatus.RETRY
            case _:
                status = ResultStatus.FAIL

        message = f"HTTP Status Code: {response.status_code}; {response.reason or response.text}"
        return Result(status, message, return_obj=return_obj)



"""
Collector plugin for DNF packages
"""
from __future__ import annotations

import fnmatch
import os
import shutil
import time

from configparser import ConfigParser
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import quote_plus, urlparse

import requests

from pydantic import SecretStr
from requests.auth import HTTPBasicAuth

import hoppr.net
import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import BatchCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_process, hoppr_rerunner
from hoppr.exceptions import HopprPluginError
from hoppr.models import HopprContext
from hoppr.models.credentials import Credentials
from hoppr.models.sbom import Component
from hoppr.models.types import PurlType, RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from packageurl import PackageURL


def _artifact_string(purl: PackageURL) -> str:
    arch = purl.qualifiers.get("arch")
    return f"{purl.name}{f'-{purl.version}' if purl.version else ''}{f'.{arch}' if arch else ''}"


def _is_rpm_repo(repo_url: str) -> bool:
    url = RepositoryUrl(url=repo_url) / "repodata/repomd.xml"
    basic_auth = None
    creds = Credentials.find(repo_url)
    if creds is not None and isinstance(creds.password, SecretStr):
        basic_auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())
    for attempt in range(3):
        if attempt > 0:
            time.sleep(5)

        resp = requests.get(f"{url}", auth=basic_auth, allow_redirects=False, stream=True, verify=True, timeout=60)
        if resp.status_code < 300:
            return True
        if resp.status_code < 500:
            return False

    return False


def _repo_proxy(url: str) -> str:
    proxy = os.getenv("https_proxy", "_none_")
    no_proxy_urls = [item for item in os.getenv("no_proxy", "").split(",") if item != ""]

    parsed_url = urlparse(url)

    for pattern in no_proxy_urls:
        if pattern in parsed_url.netloc or fnmatch.fnmatch(name=parsed_url.netloc, pat=pattern):
            proxy = "_none_"
            break

    return proxy


def _clear_cache(search_root: Path):
    """
    Clear the cache for a given Path.
    """
    temp_paths = list(search_root.rglob("hoppr-tmp-*"))
    for folder in temp_paths:
        if os.path.isdir(folder):
            shutil.rmtree(folder, ignore_errors=True)
        else:
            folder.unlink()


class CollectDnfPlugin(BatchCollectorPlugin):
    """
    Collector plugin for DNF packages
    """

    required_commands: list[str] = ["dnf"]
    supported_purl_types: list[str] = ["rpm"]
    products: list[str] = ["rpm/*"]

    def __init__(
        self,
        context: HopprContext,
        config: dict | None = None,
        config_file: str | PathLike = Path.cwd() / ".hoppr-dnf" / "dnf.conf",
    ) -> None:
        super().__init__(context=context, config=config)

        self.password_list: list[str] = []
        self.manifest_repos: list[str] = [f"{repo.url}" for repo in self.context.repositories[PurlType.RPM]]
        self.config_file = Path(config_file)

        if self.config and "dnf_command" in self.config:
            self.required_commands = [self.config["dnf_command"]]

        self.base_command = [
            self.required_commands[0],
            "--quiet",
            "--disableexcludes=all",
            f"--config={self.config_file}",
            "--disablerepo=*",
            "--enablerepo=hoppr-tmp-*",
        ]

        self.debug_args = ["--verbose"]

    def _get_component_download_url(self, purl: PackageURL) -> str:
        artifact = _artifact_string(purl)

        command = [*self.base_command, "repoquery", "--location", artifact]
        run_result = self.run_command(command, self.password_list)

        # If RPM URL not found, no need to try downloading it
        if run_result.returncode != 0 or len(run_result.stdout.decode("utf-8")) == 0:
            msg = f"{self.required_commands[0]} failed to locate package for {purl}"
            self.get_logger().debug(msg=msg, indent_level=2)

            raise HopprPluginError(msg)

        # Taking the first URL if multiple are returned
        return run_result.stdout.decode("utf-8").strip().split("\n")[0]

    def _get_found_repo(self, found_url: str) -> str | None:
        """
        Identify the repository associated with the specified URL
        """
        return next((repo for repo in self.manifest_repos if found_url.startswith(repo)), None)

    def get_version(self) -> str:
        return __version__

    @hoppr_process
    def pre_stage_process(self) -> Result:
        repo_config = ConfigParser()
        repo_config["main"] = {"cachedir": f"{self.config_file.parent / 'cache'}"}

        # Clear out cache before starting the pre_stage_process
        search_root = self.config_file.parent / "cache"
        _clear_cache(search_root)

        for idx, repo in enumerate(self.context.repositories[PurlType.RPM]):
            temp_repo = f"hoppr-tmp-{idx}"
            creds = Credentials.find(f"{repo.url}")
            if not _is_rpm_repo(f"{repo.url}"):
                self.get_logger().warning(
                    "Repo %s is not an RPM repository (repomd.xml file not found), will not be searched", repo.url
                )
                continue

            self.get_logger().debug("Creating repo %s for url %s", temp_repo, repo.url)

            # Create temporary repository file
            repo_config[temp_repo] = {
                "baseurl": f"{repo.url}",
                "enabled": "1",
                "name": f"Hoppr temp repository {idx}",
                "priority": "1",
                "proxy": _repo_proxy(f"{repo.url}"),
                "module_hotfixes": "true",
            }

            if creds is not None and isinstance(creds.password, SecretStr):
                repo_config[temp_repo]["username"] = f"{creds.username}"
                repo_config[temp_repo]["password"] = f"{creds.password.get_secret_value()}"

        if not self.context.strict_repos:
            system_repos = ConfigParser()

            # Get all system-managed .repo files
            repo_files = Path("/", "etc", "yum.repos.d").glob("*.repo")
            system_repos.read([str(repo_file) for repo_file in repo_files])

            # Get enabled system repos and add to temporary repository file
            for section in system_repos.sections():
                if system_repos[section]["enabled"] == "1":
                    repo_config.add_section(section)
                    repo_config[section] = dict(system_repos[section])

        try:
            # Create repo config dir in user directory
            config_dir = self.config_file.parent
            config_dir.mkdir(parents=True, exist_ok=True)
            with self.config_file.open(mode="w+", encoding="utf-8") as repo_file:
                repo_config.write(repo_file, space_around_delimiters=False)
        except OSError as ex:
            return Result.fail(f"Unable to write DNF repository config file: {ex}")

        # Populate user DNF cache
        command = [
            *self.base_command,
            "check-update",
            "makecache",
            *(self.debug_args if self.get_logger().is_verbose() else []),
        ]

        # Generate cache to use when downloading components
        result = self.run_command(command, self.password_list)
        self.get_logger().debug(msg=(f"Return Code: {result.returncode}"), indent_level=2)
        if result.returncode != 0:
            if hoppr.utils.get_partition(config_dir).fstype == "vboxsf":
                self.get_logger().warning(
                    "NOTE: Possible attempt to run DNF from a shared vagrant directory('%s'). "
                    "Recommend retrying outside of the shared vagrant directory",
                    config_dir,
                )
            return Result.fail(message="Failed to populate DNF cache.")

        return Result.success()

    @hoppr_rerunner
    def collect(self, comp: Component) -> Result:
        """
        Copy a component to the local collection directory structure
        """

        purl = hoppr.utils.get_package_url(comp.purl)

        self.get_logger().info(msg=f"Copying DNF package from {purl}", indent_level=2)

        # Try getting RPM URL
        try:
            found_url = self._get_component_download_url(purl)
        except HopprPluginError as ex:
            return Result.retry(message=str(ex))

        repo = self._get_found_repo(found_url)
        # Return failure if found RPM URL is not from a repo defined in the manifest
        if repo is None:
            return Result.fail(
                "Successfully found RPM file but URL does not match any repository in manifest."
                f" (Found URL: '{found_url}')"
            )

        result = self.check_purl_specified_url(purl, repo)
        if not result.is_success():
            return result

        found_url_path = Path(found_url)
        subdir = found_url_path.relative_to(repo).parent
        target_dir = self.directory_for(purl.type, repo, subdir=str(subdir))

        # Download the RPM file to the new directory
        response = hoppr.net.download_file(
            url=found_url, dest=str(target_dir / found_url_path.name), creds=Credentials.find(url=repo)
        )
        result = Result.from_http_response(response)

        if result.is_retry():
            msg = f"Failed to download DNF artifact {purl.name} version {purl.version} from {found_url}"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.retry(message=msg)

        if result.is_success():
            self.set_collection_params(comp, repo, target_dir)
            result = Result.success(result.message, return_obj=comp)

        return result

    @hoppr_process
    def post_stage_process(self) -> Result:
        # Find repodata folders created when cache was generated
        search_root = self.config_file.parent / "cache"
        paths = list(search_root.rglob("hoppr-tmp-*/repodata"))

        repo_config = ConfigParser()
        repo_config.read(filenames=self.config_file, encoding="utf-8")

        msg = "Copying DNF repodata into bundle"
        self.get_logger().info(msg)
        self.notify(msg, type(self).__name__)

        # Loop over repos defined in the temp dnf.conf file
        for section in repo_config.sections():
            # Loop over repodata folders in DNF cache
            for folder in paths:
                # Check if full repodata path contains temp repo name
                if section in str(folder):
                    # Get temp repo's URL as defined in manifest
                    repo_url = repo_config[section]["baseurl"]
                    target_dir = self.context.collect_root_dir / "rpm" / quote_plus(repo_url) / "repodata"

                    shutil.copytree(src=folder, dst=target_dir, dirs_exist_ok=True)

        msg = f"Removing config file {self.config_file}"
        self.get_logger().info(msg)
        self.notify(msg, type(self).__name__)

        if self.config_file.exists():
            try:
                self.config_file.unlink()
            except FileNotFoundError as ex:
                return Result.fail(f"Failed to remove temporary DNF config file: {ex}")

        msg = "Removing cached DNF repodata"
        self.get_logger().info(msg)
        self.notify(msg, type(self).__name__)

        # Clear cache at the end of the post stage process
        _clear_cache(search_root)

        return Result.success()



"""
A logger that gets dumped to stdout when closed
"""
from __future__ import annotations

import functools
import inspect
import itertools
import logging

from logging import FileHandler, Formatter, Logger, LoggerAdapter
from logging.handlers import MemoryHandler
from threading import _RLock as RLock
from typing import Callable, MutableMapping, cast


def locked(func: Callable) -> Callable:
    """
    Acquire logfile lock, run the wrapped function, then release the lock
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.lock:
            self.lock.acquire()

        func(self, *args, **kwargs)

        if self.lock:
            self.lock.release()

    return wrapper


class HopprLogger(LoggerAdapter):
    """
    Logger that buffers log records in memory until flushed
    """

    id_iter = itertools.count()

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str | None = None,
        filename: str = "hoppr.log",
        level: int = logging.INFO,
        lock: RLock | None = None,
        flush_immed: bool = False,
    ) -> None:
        """
        Initialize the logger, formatter, and handlers

        Args:
            name (str | None, optional): Name to assign the logger. Defaults to None.
            filename (str, optional): Log file to create. Defaults to "hoppr.log".
            level (int, optional): Logging level. Defaults to logging.INFO.
            lock (RLock | None, optional): Lock object to acquire/release when writing to log file. Defaults to None.
            flush_immed (bool, optional): Flush immediately when logging. Defaults to False.
        """
        self.extra: MutableMapping[str, object]
        self.logger: Logger

        self.filename = filename
        self.flush_immed: bool = flush_immed
        self.instance_id: int = next(self.id_iter)
        self.lock: RLock | None = lock

        _, frame_info, *_ = inspect.stack(context=2)
        caller = frame_info.function

        if (caller_cls := frame_info.frame.f_locals.get("self")) is not None:
            caller = type(caller_cls).__name__

        extra = cast(MutableMapping, {})
        extra["caller"] = caller

        name = f"{name}-{self.instance_id}" if name else f"{caller}-{self.instance_id}"

        logger = logging.getLogger(name)
        logger.setLevel(level)

        formatter = Formatter(
            fmt="[$asctime] - [${caller}] - [$levelname] - $message",
            style="$",
            defaults={"caller": caller},
        )

        file_handler = FileHandler(filename)
        file_handler.setFormatter(formatter)

        log_handler = MemoryHandler(10000, flushLevel=logging.CRITICAL, target=file_handler)
        logger.addHandler(log_handler)

        super().__init__(logger, extra)

    def clear_targets(self) -> None:
        """
        Set target for all MemoryHandlers in this logger to None

        Thus when these handlers are flushed, nothing will go to standard output
        """
        for handler in self.logger.handlers:
            if isinstance(handler, MemoryHandler):
                handler.setTarget(None)

    @locked
    def close(self) -> None:
        """
        Close (and flush) all handlers for this logger
        """
        for handler in self.logger.handlers:
            handler.close()

    @locked
    def flush(self) -> None:
        """
        Flush all handlers for this logger
        """
        for handler in self.logger.handlers:
            handler.flush()

    def is_verbose(self) -> bool:
        """
        Check whether running with `--debug`/`--verbose` flag
        """
        return self.logger.level == logging.DEBUG

    def log(self, level: int, msg: object, *args, **kwargs) -> None:
        """
        Wrapper function for logging messages
        """
        super().log(level, msg, *args, **kwargs)

        if self.flush_immed:
            self.flush()

    def process(self, msg: str, kwargs: MutableMapping[str, object]) -> tuple[str, MutableMapping[str, object]]:
        indent_level = cast(int, kwargs.pop("indent_level", 0))
        indent_string = " " * 4 * indent_level
        msg = f"{indent_string}{msg}"

        return super().process(msg, kwargs)



"""
Plugin to cascade through various collectors to try to capture an artifact
"""
from __future__ import annotations

import logging

from typing import Any

from hoppr import __version__, constants
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_ignore_excluded, hoppr_process
from hoppr.exceptions import HopprPluginError
from hoppr.models import HopprContext
from hoppr.models.transfer import ComponentCoverage
from hoppr.result import Result
from hoppr.utils import plugin_class, plugin_instance


class CompositeCollector(HopprPlugin):
    """
    Plugin to cascade through various collectors to try to capture an artifact
    """

    default_component_coverage = ComponentCoverage.EXACTLY_ONCE

    def get_version(self) -> str:
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        if self.config is None:
            raise HopprPluginError("CompositeCollector requires configuration be defined in the transfer config file")

        if (
            constants.ConfigKeys.PLUGINS not in self.config
            or not isinstance(self.config[constants.ConfigKeys.PLUGINS], list)
            or len(self.config[constants.ConfigKeys.PLUGINS]) == 0
        ):
            raise HopprPluginError("CompositeCollector requires at least one plugin be configured")

        self.child_plugins: list[HopprPlugin] = []

        for plugin_spec in self.config[constants.ConfigKeys.PLUGINS]:
            plugin = plugin_instance(plugin_spec["name"], context, plugin_spec.get(constants.ConfigKeys.CONFIG))
            self.child_plugins.append(plugin)

    @hoppr_process
    def pre_stage_process(self):
        """
        Run all child plugin pre-stage processes
        """

        for plugin in self.child_plugins:
            result = plugin.pre_stage_process()
            if result.is_fail() or result.is_retry():
                return Result.fail(f"Failure running {type(plugin).__name__}")

        return Result.success()

    @hoppr_ignore_excluded
    @hoppr_process
    def process_component(self, comp: Any) -> Result:  # pylint: disable=unused-argument
        """
        Run component through each child plugin component processes
        """

        for plugin in self.child_plugins:
            self._log_and_flush("Attempting to process %s using %s", comp.purl, type(plugin).__name__)

            result = plugin.process_component(comp)
            if result.is_success():
                return Result.success(f"Used {type(plugin).__name__}")

        return Result.fail(f"Failed to run component {comp.purl} through all child plugins.")

    @hoppr_process
    def post_stage_process(self):
        """
        Run all child plugin post-stage processes
        """
        for plugin in self.child_plugins:
            msg = f"Running post stage process for {type(plugin).__name__}"
            self.get_logger().info(msg)
            self.notify(msg, type(self).__name__)

            result = plugin.post_stage_process()

            if result.is_fail() or result.is_retry():
                return Result.fail(f"Failure running {type(plugin).__name__}")

        return Result.success()

    def _log_and_flush(self, msg: str, *args, level: int = logging.INFO, indent_level: int = 0):
        self.get_logger().log(level, msg, *args, indent_level=indent_level)
        self.get_logger().flush()

    def supports_purl_type(self, purl_type: str) -> bool:
        """
        Composite collector purl type support is based on first child collector
        """
        return self.child_plugins[0].supports_purl_type(purl_type)

    @classmethod
    def get_attestation_products(cls, config: dict | None = None) -> list[str]:
        """
        Return a list of attestation products for this class
        """

        if (
            config is None
            or constants.ConfigKeys.PLUGINS not in config
            or not isinstance(config[constants.ConfigKeys.PLUGINS], list)
            or len(config[constants.ConfigKeys.PLUGINS]) == 0
        ):
            raise HopprPluginError("CompositeCollector requires at least one plugin be configured")

        plugin_cls = plugin_class(config[constants.ConfigKeys.PLUGINS][0]["name"])
        plugin_config = config[constants.ConfigKeys.PLUGINS][0].get(constants.ConfigKeys.CONFIG)

        return plugin_cls.get_attestation_products(plugin_config)



"""
Models to express `licenses` fields
"""
from __future__ import annotations

from typing import Annotated, ClassVar

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from hoppr_cyclonedx_models import spdx
from pydantic import Extra, Field

from hoppr.models.base import CycloneDXBaseModel, UniqueIDMap


class License(CycloneDXBaseModel, cdx.License):
    """
    License data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for License model"
        extra = Extra.allow

    unique_id_map: ClassVar[UniqueIDMap] = {}


class LicenseIDRequired(License):
    """
    License data model derived from CycloneDXBaseModel
    """

    # Override to make `id` field required by removing default value
    id: Annotated[
        spdx.LicenseID | None,
        Field(description="A valid SPDX license ID", examples=["Apache-2.0"], title="License ID (SPDX)"),
    ]

    unique_id_map: ClassVar[UniqueIDMap] = {}


class LicenseNameRequired(License):
    """
    License data model derived from CycloneDXBaseModel
    """

    # Override to make `name` field required by removing default value
    name: Annotated[
        str | None,
        Field(
            description="If SPDX does not define the license used, this field may be used to provide the license name",
            examples=["Acme Software License"],
            title="License Name",
        ),
    ]

    unique_id_map: ClassVar[UniqueIDMap] = {}


class LicenseMultipleItem(CycloneDXBaseModel):  # pylint: disable=missing-class-docstring
    license: Annotated[LicenseIDRequired | LicenseNameRequired, Field()]

    unique_id_map: ClassVar[UniqueIDMap] = {}


class LicenseExpressionItem(CycloneDXBaseModel):
    """
    SPDX license expression data model
    """

    expression: Annotated[
        str,
        Field(
            title="SPDX License Expression",
            examples=["Apache-2.0 AND (MIT OR GPL-2.0-only)", "GPL-3.0-only WITH Classpath-exception-2.0"],
        ),
    ]
    bom_ref: Annotated[
        str | None,
        Field(
            title="BOM Reference",
            description=(
                "An optional identifier which can be used to reference the license elsewhere in the BOM. "
                "Every bom-ref MUST be unique within the BOM."
            ),
            alias="bom-ref",
            min_length=1,
        ),
    ] = None

    unique_id_map: ClassVar[UniqueIDMap] = {}


LicenseExpression = Annotated[
    list[LicenseExpressionItem],
    Field(
        title="SPDX License Expression",
        description="A tuple of exactly one SPDX License Expression.",
        min_items=1,
        max_items=1,
    ),
]


LicenseChoice = list[LicenseMultipleItem] | LicenseExpression | None



"""
Run Hoppr processing, using multiple processors
"""
from __future__ import annotations

import importlib
import logging
import os
import shutil
import socket
import sys
import tempfile
import time
import uuid

from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from datetime import datetime
from multiprocessing import cpu_count
from os import PathLike
from pathlib import Path
from threading import _RLock as RLock
from threading import get_ident
from typing import TYPE_CHECKING, Any, MutableMapping
from urllib.parse import quote_plus

import jmespath

from rich.live import Live

import hoppr.utils

from hoppr.cli.layout import console
from hoppr.core_plugins.report_generator import Report, ReportGenerator
from hoppr.exceptions import HopprPluginError
from hoppr.in_toto import HopprInTotoLinks
from hoppr.logger import HopprLogger
from hoppr.models import HopprContext
from hoppr.models.credentials import Credentials
from hoppr.models.manifest import Manifest, ManifestFile
from hoppr.models.sbom import Component, Sbom
from hoppr.models.transfer import ComponentCoverage, Plugin, StageRef, Transfer
from hoppr.models.types import BomAccess
from hoppr.result import Result

# Import on Unix systems only
if os.name == "posix":
    import pwd

if TYPE_CHECKING:
    from hoppr.cli.bundle import HopprBundleLayout, HopprBundleSummary

layout: HopprBundleLayout
summary_panel: HopprBundleSummary


def _run_plugin(
    plugin_name: str,
    context: HopprContext,
    config: dict[str, Any] | None,
    method_name: str,
    component: Component | None,
) -> Result:
    """
    Runs a single method for a single component (if supplied) on a single plugin
    """
    plugin = hoppr.utils.plugin_instance(plugin_name, context, config)
    plugin.subscribe(observer=layout, callback=layout.print)  # pyright: ignore

    match method_name:
        case "pre_stage_process":
            result = plugin.pre_stage_process()
        case "process_component":
            result = plugin.process_component(component)
        case "post_stage_process":
            result = plugin.post_stage_process()
        case _:
            result = Result.fail(f"Invalid method: {method_name}")

    if result.return_obj is not None and not plugin.bom_access.has_access_to(result.return_obj):
        result = Result.fail(
            f"Plugin {type(plugin).__name__} has BOM access level {plugin.bom_access.name}, "
            f"but returned an object of type {type(result.return_obj).__name__}"
        )

    plugin.unsubscribe(observer=layout)  # pyright: ignore

    return result


class StageProcessor:  # pylint: disable=too-few-public-methods
    """
    Class to handle all processing within a single Hoppr stage
    """

    component_based_methods = ["process_component"]

    def __init__(self, stage_ref: StageRef, context: HopprContext, logger: HopprLogger):
        self.context = context
        self.logger = logger
        self.plugin_ref_list = hoppr.utils.dedup_list(stage_ref.plugins)
        self.required_coverage = ComponentCoverage.OPTIONAL
        self.results: dict[str, list[tuple[str, str | None, Result]]] = {}
        self.stage_id = stage_ref.name

        self.config_component_coverage = None
        if stage_ref.component_coverage is not None:
            self.config_component_coverage = ComponentCoverage[str(stage_ref.component_coverage)]

    def run(self) -> Result:
        """
        Run all processes for this stage
        """
        try:
            self.plugin_ref_list = self._get_stage_plugins()
            self.required_coverage = self._get_required_coverage()

            if not self.plugin_ref_list and str(self.required_coverage) in {"EXACTLY_ONCE", "AT_LEAST_ONCE"}:
                raise HopprPluginError(f"No plugins were loaded, but required coverage is {self.required_coverage}")
        except (ModuleNotFoundError, HopprPluginError) as err:
            return Result.fail(str(err))

        result = self._check_bom_access()
        if not result.is_success():
            return result

        # Run each sub-stage (pre_stage_process, process_component, post_stage_process)
        # for all plugins (and, for process_component, for all components).
        # Each sub-stage must complete before the next can begin
        output_width = console.width if hoppr.utils.is_basic_terminal() else layout.output_panel.get_width()

        plugins = [hoppr.utils.plugin_class(plugin.name) for plugin in self.plugin_ref_list]

        layout.overall_progress.title = f"[bold blue]Stage: {self.stage_id}"

        # Increment overall progress bar total by 2 to account for pre and post processing steps
        layout.progress_task.total = (layout.progress_task.total or 0) + 2

        def _method_label(method_name: str, label: str):
            "Print label for method if the named method is defined directly in the plugin class"
            if any(method_name in plugin.__dict__ for plugin in plugins):
                layout.progress_task.description = label.capitalize()
                layout.print("", source="")
                layout.print(f" Stage: {self.stage_id} ({label}) ".center(output_width, "="))

        _method_label("pre_stage_process", "pre-processing")
        result.merge(self._run_all("pre_stage_process"))
        layout.overall_progress.progress_bar.advance(layout.progress_task.id)

        _method_label("collect", "processing")
        result.merge(self._run_all("process_component"))

        _method_label("post_stage_process", "post-processing")
        result.merge(self._run_all("post_stage_process"))
        layout.overall_progress.progress_bar.advance(layout.progress_task.id)

        return result

    def _populate_futures(self, method_name: str) -> tuple[int, int]:
        """
        Run the named method for all plugins.

        If appropriate to the method, run it for all components for all plug-ins.
        """
        future_argument_map: dict[Future[Result], tuple[Plugin, Component | None]] = {}

        def _components():
            yield from self.context.delivered_sbom.components or []

        with ThreadPoolExecutor(max_workers=self.context.max_processes, thread_name_prefix=self.stage_id) as executor:
            failures = 0
            retries = 0

            # Prevent circular import
            base_cls = getattr(importlib.import_module(name="hoppr.base_plugins.hoppr"), "HopprPlugin")

            for plugin in self.plugin_ref_list:
                plugin_cls = hoppr.utils.plugin_class(plugin.name)

                # Skip plugin if method not defined in class directly
                if getattr(plugin_cls, method_name) == getattr(base_cls, method_name):
                    continue

                if method_name in self.component_based_methods:
                    layout.start_job(plugin_cls.__name__)

                    # Create one concurrent future object to run this method for each component
                    for component in _components():
                        if component.purl:
                            instance = hoppr.utils.plugin_instance(plugin.name, self.context, plugin.config)

                            if instance.supports_purl_type(hoppr.utils.get_package_url(component.purl).type):
                                future_proc = executor.submit(
                                    _run_plugin,
                                    plugin_name=plugin.name,
                                    context=self.context,
                                    config=plugin.config,
                                    method_name=method_name,
                                    component=component,
                                )

                                future_argument_map[future_proc] = (plugin, component)

                    layout.stop_job(plugin_cls.__name__)
                else:
                    # Create a concurrent future object to run this method
                    future_proc = executor.submit(
                        _run_plugin,
                        plugin_name=plugin.name,
                        context=self.context,
                        config=plugin.config,
                        method_name=method_name,
                        component=None,
                    )

                    future_argument_map[future_proc] = (plugin, None)

            fail, retry = self.process_status(
                future_argument_map=future_argument_map,
                method_name=method_name,
                failures=failures,
                retries=retries,
            )

            failures += fail
            retries += retry

        return failures, retries

    def process_status(
        self,
        future_argument_map: dict[Future[Result], tuple[Plugin, Component | None]],
        method_name: str,
        failures: int,
        retries: int,
    ) -> tuple[int, int]:
        """
        Print processing results.
        """
        for future in as_completed(future_argument_map):
            # Retrieve the result
            plugin, comp = future_argument_map[future]
            future_result: Result = future.result()

            if not future_result.is_skip():
                plugin_cls = hoppr.utils.plugin_class(plugin.name)

                if not future_result.is_excluded():
                    self._save_result(method_name, plugin_cls.__name__, future_result, comp)
                    self._update_bom(future_result.return_obj, comp)

                if method_name in self.component_based_methods:
                    layout.advance_job(plugin_cls.__name__)

                self._report_result(plugin_cls.__name__, comp, future_result)

                ReportGenerator.report_gen_list.append(
                    Report(
                        uuid.uuid4(),
                        plugin_cls.__name__,
                        self.stage_id,
                        future_result.status,
                        future_result.message,
                        method_name,
                        comp,
                    )
                )

                if future_result.is_fail():
                    layout.update_job(plugin_cls.__name__, status="fail")
                    failures += 1

                if future_result.is_retry():
                    retries += 1

        return failures, retries

    def _run_all(self, method_name: str):
        """
        Run the named method for all plugins and process results
        """
        failures, retries = self._populate_futures(method_name)

        if method_name in self.component_based_methods:
            failures += self._check_component_coverage(method_name)

        return self._get_stage_result(method_name, failures, retries)

    def _get_stage_result(self, method_name: str, failures: int, retries: int) -> Result:
        match (failures, retries):
            case (0, 0):
                result = Result.success()
            case (0, retry) if retry > 0:
                result = Result.fail(f"{retries} '{method_name}' processes returned 'retry'")
            case (fail, 0) if fail > 0:
                result = Result.fail(f"{failures} '{method_name}' processes failed")
            case _:
                result = Result.fail(f"{failures} '{method_name}' processes failed, and {retries} returned 'retry'")

        return result

    def _get_required_coverage(self) -> ComponentCoverage:
        if self.config_component_coverage is not None:
            return self.config_component_coverage

        if len(self.plugin_ref_list) == 0:
            return ComponentCoverage.OPTIONAL

        plugin = hoppr.utils.plugin_class(self.plugin_ref_list[0].name)
        coverage = plugin.default_component_coverage

        for plugin_ref in self.plugin_ref_list:
            plugin = hoppr.utils.plugin_class(plugin_ref.name)
            if plugin.default_component_coverage != coverage:
                raise HopprPluginError(
                    f"Plugins for stage {self.stage_id} do not have consistent default "
                    "component coverage values. The value may be overridden in transfer file."
                )

        return coverage

    def _check_component_coverage(self, method_name: str) -> int:
        result_count: dict[str | None, int] = {}

        for _, bom_ref, _ in self.results.get(method_name, []):
            result_count[bom_ref] = result_count.get(bom_ref, 0) + 1

        additional_failures = 0
        for component in self.context.delivered_sbom.components or []:
            count = result_count.get(component.purl, 0)
            if not self.required_coverage.accepts_count(count) and str(component.scope) not in {
                "excluded",
                "Scope.excluded",
            }:
                bad_comp_result = Result.fail(
                    f"Component processed {count} times, {self.required_coverage.name} coverage required"
                )
                self._save_result(method_name, f"Stage {self.stage_id}", bad_comp_result, component)
                self._report_result(f"Stage {self.stage_id}", component, bad_comp_result)
                additional_failures += 1

        return additional_failures

    def _check_bom_access(self) -> Result:
        access_counts: dict[BomAccess, list[str]] = {access: [] for access in BomAccess}

        for plugin_ref in self.plugin_ref_list:
            plugin_cls = hoppr.utils.plugin_class(plugin_ref.name)
            access_counts[plugin_cls.bom_access].append(plugin_cls.__name__)

        if len(access_counts[BomAccess.FULL_ACCESS]) > 0 and len(self.plugin_ref_list) > 1:
            msg = (
                f"Stage {self.stage_id} has one or more plugins with {BomAccess.FULL_ACCESS.name}: "
                f"{', '.join(access_counts[BomAccess.FULL_ACCESS])}"
                ", and multiple plugins defined for the stage."
                "\n    Any plugin with FULL BOM access must be the only plugin in the stage"
            )
            layout.print(msg, style="red")
            return Result.fail(msg)

        if len(access_counts[BomAccess.COMPONENT_ACCESS]) > 0 and self.required_coverage.max_value > 1:
            msg = (
                f"Stage {self.stage_id} has one or more plugins with {BomAccess.COMPONENT_ACCESS.name}: "
                f"{', '.join(access_counts[BomAccess.COMPONENT_ACCESS])}"
                f", and required component coverage for the stage of {self.required_coverage.name}."
                "\n    If any plugins have COMPONENT access, the stage required coverage must be "
                "EXACTLY_ONCE or NO_MORE_THAN_ONCE."
            )
            layout.print(msg, style="red")
            return Result.fail(msg)

        return Result.success()

    def _update_bom(self, return_obj: Any, comp: Component | None):
        if self.context.delivered_sbom.components is not None and isinstance(return_obj, Component):
            for index, delivered_comp in enumerate(self.context.delivered_sbom.components):
                if delivered_comp == comp:
                    self.context.delivered_sbom.components[index] = return_obj
                    break

        elif isinstance(return_obj, Sbom):
            self.context.delivered_sbom = return_obj

    @staticmethod
    def _report_result(plugin: str, comp: Component | None, result: Result):
        if result.is_success():
            color = "green"
        elif result.is_excluded():
            color = "yellow"
        else:
            color = "red"

        desc = f"[bold {color}]{result.status.name}[/]"

        if result.message:
            desc = f"{desc}: {result.message}"

        if comp is not None:
            desc = f"{desc} for {comp.purl}"

        layout.print(desc, source=plugin)

    def _save_result(self, method_name: str, plugin: str, result: Result, comp: Component | None):
        """
        Store the results for later use
        """
        comp_string = comp.bom_ref if comp is not None else None

        # If needed, create a new list for this method
        # Might need to expand this definition in the future to separate by plug-in
        if method_name not in self.results:
            self.results[method_name] = []

        self.results[method_name].append((plugin, comp_string, result))

    def _get_stage_plugins(self) -> list[Plugin]:
        """
        Determine list of plugin references used in this stage
        """
        used_purl_types: set[str] = set()

        # Get list of all component PURLs that are not None
        results = jmespath.search(
            expression="components[*].purl | not_null(@)",
            data=self.context.consolidated_sbom.dict(),
        )

        used_purl_types.update(hoppr.utils.get_package_url(purl).type for purl in results)

        plugin_list: list[Plugin] = []

        with (Path(self.context.collect_root_dir) / "generic" / "_metadata_" / "_run_data_").open(
            mode="a", encoding="utf-8"
        ) as rundata:
            self.logger.info(f"Plugins for stage {self.stage_id}:")
            rundata.write(f"\nPlugins for stage {self.stage_id}:\n")

            for plugin_ref in self.plugin_ref_list:
                instance = hoppr.utils.plugin_instance(plugin_ref.name, self.context, plugin_ref.config)

                # Determine if plugin's `supported_purl_types` are in the set of PURL types defined in SBOM components
                if used_purl_types.intersection(instance.supported_purl_types):
                    plugin_list.append(plugin_ref)

                    msg = f"{type(instance).__name__} version {instance.get_version()} from {plugin_ref.name}"
                    self.logger.info(msg, indent_level=1)
                    rundata.write(f"    {msg}\n")

                    if "collect" in type(instance).__dict__:
                        # Get all components in input SBOM supported by the plugin and add job to side panel
                        layout.add_job(
                            description=type(instance).__name__,
                            total=sum(
                                len(jmespath.search(expression=f"[? starts_with(@, 'pkg:{supported}')]", data=results))
                                for supported in instance.supported_purl_types
                            ),
                        )

        return plugin_list


class HopprProcessor:  # pylint: disable=too-many-instance-attributes
    """
    Run the Hoppr process
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        transfer_file: Path,
        manifest_file: Path,
        credentials_file: Path | None,
        create_attestations: bool = False,
        functionary_key_path: Path | None = None,
        functionary_key_password: str | None = None,
        log_level: int = logging.INFO,
        log_file: str | Path | None = None,
        strict_repos: bool = True,
        previous_delivery: Path | None = None,
    ) -> None:
        self.context: HopprContext
        self.live_display = Live(renderable=layout, console=console, refresh_per_second=10)
        self.log_file: Path = Path(log_file or f"hoppr_{time.strftime('%Y%m%d-%H%M%S')}.log")
        self.log_level = log_level
        self.logfile_lock = RLock()
        self.metadata_files: list[Path] = [manifest_file, transfer_file]
        self.previous_delivery = previous_delivery
        self.strict_repos: bool = strict_repos

        if not hoppr.utils.is_basic_terminal():
            self.live_display.start(refresh=True)

        self.logger = self.get_logger(
            log_name=type(self).__name__,
            log_file=self.log_file,
            log_level=self.log_level,
        )

        self.logger.info("Loading %s and %s files...", manifest_file.name, transfer_file.name)
        self.manifest_file = ManifestFile.parse_file(manifest_file)
        self.manifest = Manifest.load(manifest_file)
        self.transfer = Transfer.load(transfer_file)

        self.credentials = None
        if credentials_file is not None:
            self.logger.info("Loading %s file...", credentials_file.name)
            self.credentials = Credentials.load(credentials_file)
            self.metadata_files.append(credentials_file)

        layout.print(f"Loading {manifest_file.name} and {transfer_file.name} files...")
        self.manifest_file = ManifestFile.parse_file(manifest_file)
        self.manifest = Manifest.load(manifest_file)
        self.transfer = Transfer.load(transfer_file)

        self.stage_processor_map: MutableMapping[StageRef, StageProcessor] = {}

        self.in_toto_links = HopprInTotoLinks(
            create_attestations,
            self.transfer,
            functionary_key_path,
            functionary_key_password,
        )

        self.report_gen: ReportGenerator

    def _collect_file(
        self, file_name: str | PathLike[str], target_dir: str | PathLike[str]
    ) -> None:  # pragma: no cover
        self.get_logger().info("Collecting metadata file %s", file_name)
        abs_path = Path(file_name).absolute()

        target = Path(target_dir, quote_plus(f"{abs_path}"))
        shutil.copyfile(file_name, f"{target}")

    def _collect_manifest_metadata(self, manifest: ManifestFile, target_dir: str | PathLike[str]) -> None:
        layout.print(f"Loading {len(manifest.includes)} includes from manifest...")

        for include_ref in manifest.includes:
            include = Manifest.loaded_manifests[include_ref]
            include_dict = include_ref.dict()

            if "local" in include_dict:
                url = str(Path(include_dict["local"]).resolve())
            else:
                url = str(include_dict["url"])

            target_file = Path(target_dir) / quote_plus(url)

            with target_file.open(mode="w+", encoding="utf-8") as output_file:
                output_file.write(include.yaml(by_alias=True, exclude_unset=True, indent=True))

        layout.print(f"Loading {len(manifest.sboms)} sboms from manifest...")

        for sbom_ref in manifest.sboms:
            sbom = Sbom.loaded_sboms[sbom_ref]
            sbom_dict = sbom_ref.dict()

            if "local" in sbom_dict:
                url = str(Path(sbom_dict["local"]).resolve())
            elif "oci" in sbom_dict:
                url = str(sbom_dict["oci"])
            else:
                url = str(sbom_dict["url"])

            target_file = Path(target_dir) / quote_plus(url)

            with target_file.open(mode="w+", encoding="utf-8") as output_file:
                output_file.write(sbom.json(by_alias=True, exclude_unset=True, indent=2))

    def _collect_metadata(self):
        layout.print("Collecting Hoppr Metadata")
        target_dir = Path(self.context.collect_root_dir) / "generic" / "_metadata_"
        target_dir.mkdir(exist_ok=True, parents=True)

        if os.name == "posix":
            user = pwd.getpwuid(os.getuid()).pw_name
        else:
            user = os.getlogin()  # pragma: no cover

        with (target_dir / "_run_data_").open(mode="w", encoding="utf-8") as rundata:
            rundata.write(
                f"Hoppr Version:     {hoppr.__version__}\n"
                f"Collection Start:  {str(datetime.now())}\n"
                f"User:              {user}\n"
                f"Host FQDN:         {socket.getfqdn()}\n"
                f"Working directory: {os.getcwd()}\n"
                f"Argument List:     {sys.argv[1:]}\n"
            )

        for file_name in self.metadata_files:
            self._collect_file(file_name, target_dir)

        self._collect_manifest_metadata(self.manifest_file, target_dir)

    def _collect_consolidated_bom(self) -> None:
        target_dir = Path(self.context.collect_root_dir) / "generic" / "_metadata_"
        target_dir.mkdir(parents=True, exist_ok=True)

        with (target_dir / "_consolidated_bom.json").open(mode="w", encoding="utf-8") as bom_data:
            bom_data.write(self.context.consolidated_sbom.json(exclude_none=True, by_alias=True, indent=2))

    def _collect_delivered_bom(self, stage_name: str | None = None) -> None:
        target_dir = Path(self.context.collect_root_dir) / "generic" / "_metadata_"
        target_dir.mkdir(parents=True, exist_ok=True)

        preface = f"intermediate_{stage_name}_" if stage_name else ""

        with (target_dir / f"_{preface}delivered_bom.json").open(mode="w", encoding="utf-8") as bom_data:
            bom_data.write(self.context.delivered_sbom.json(exclude_none=True, by_alias=True, indent=2))

    def _summarize_results(self) -> int:
        """
        Summarize the results of a HopprProcess run
        """
        for stage_ref, stage in self.stage_processor_map.items():
            stage_summary = summary_panel.add_stage_result(stage_ref.name)

            for method_name, result_list in stage.results.items():
                result_count = len(result_list)
                summary_panel.total_success_count += result_count
                failure_count = 0

                for plugin_name, comp_str, result in result_list:
                    # All retries should be handled internally by the plugins,
                    # So if a RETRY result is returned, that's a failure
                    if result.is_fail() or result.is_retry():
                        failure_count += 1
                        summary_panel.add_failure(plugin_name, comp_str, result.message)

                summary_panel.add_method_result(stage_ref.name, method_name, result_count, failure_count)

            summary_panel.stage_results_map[stage_ref.name].add_row()
            summary_panel.summary_group.renderables.append(stage_summary)

        console.print("\n", summary_panel)

        self.report_gen = ReportGenerator(self.context)
        self.report_gen.generate_report()

        return summary_panel.total_failure_count

    def get_logger(
        self,
        log_name: str | None = None,
        log_file: str | Path = "hoppr.log",
        log_level: int = logging.INFO,
    ) -> HopprLogger:
        """
        Returns the logger for this class

        Args:
            log_name (str | None, optional): Name of logger. Defaults to None.
            log_file (str | Path | None, optional): Path to log file. Defaults to "hoppr.log".
            log_level (int, optional): Logging level. Defaults to logging.INFO.

        Returns:
            HopprLogger: A new HopprLogger instance.
        """
        if not hasattr(self, "logger"):
            self.logger = HopprLogger(
                filename=str(log_file),
                name=log_name or f"HopprProcessor--{get_ident()}",
                level=log_level,
                flush_immed=True,
            )

        return self.logger

    def run(self) -> Result:
        """
        Run the Hoppr process executing each stage in turn

        Args:
            log_file (Path | None, optional): Path to log file. Defaults to None.
            strict_repos (bool, optional): Enables strict repository searches for components. Defaults to True.

        Returns:
            Result: The composite result of all executed stages
        """
        result = Result.success()

        with tempfile.TemporaryDirectory() as collection_root:
            self.context = HopprContext(
                collect_root_dir=Path(collection_root).resolve(),
                consolidated_sbom=self.manifest.consolidated_sbom.copy(deep=True),
                credential_required_services=getattr(self.credentials, "credential_required_services", None),
                delivered_sbom=self.manifest.consolidated_sbom.copy(deep=True),
                log_level=self.log_level,
                logfile_location=self.log_file,
                logfile_lock=self.logfile_lock,
                max_processes=self.transfer.max_processes or cpu_count(),
                repositories=self.manifest.repositories,
                sboms=list(Sbom.loaded_sboms.values()),
                stages=self.transfer.stages,
                strict_repos=self.strict_repos,
                previous_delivery=self.previous_delivery,
            )

            if self.context.consolidated_sbom is not None and len(self.context.consolidated_sbom.components or []) == 0:
                msg = "No SBOMs defined in manifests, or SBOMs contain no components. Nothing to process."
                layout.print(msg, style="red")
                return Result.fail(msg)

            self.in_toto_links.set_collection_root(collection_root)
            self.in_toto_links.record_stage_start("_collect_metadata")
            self._collect_metadata()
            self._collect_consolidated_bom()
            self.in_toto_links.record_stage_stop("_collect_metadata")

            # Reset some class variables (we should refactor to use instance variables, see issue
            # https://gitlab.com/hoppr/hoppr/-/issues/230) so that:
            #   1: The consolidated bom cannot be modified
            #   2: Plug-ins may choose to load other manifests.
            Sbom.unique_id_map = {}
            Component.unique_id_map = {}
            Manifest.loaded_manifests = {}

            msg = f"Beginning Hoppr Process execution, max_processes={self.context.max_processes}"
            self.logger.info(msg=msg)
            layout.print(msg)

            for stage_ref in self.transfer.stages:
                self.logger.info(f" Beginning Stage {stage_ref.name} ".center(100, "="))

                stage = StageProcessor(stage_ref, self.context, self.logger)
                self.in_toto_links.record_stage_start(stage_ref.name)
                self.stage_processor_map[stage_ref] = stage

                result = stage.run()

                if result.is_fail() or result.is_retry():
                    msg = f"Stage {stage_ref.name} failed, processing terminated: {result.message}"
                    self.logger.error(msg=msg)
                    layout.print(f"  {msg}", style="red")

                    break

                self._collect_delivered_bom(stage_ref.name)
                self._collect_delivered_bom()
                self.in_toto_links.record_stage_stop(stage_ref.name)

            self.in_toto_links.record_stage_start("_finalize")
            self._collect_delivered_bom()
            self.in_toto_links.record_stage_stop("_finalize")

        if not hoppr.utils.is_basic_terminal():
            self.live_display.stop()

        failed_jobs = self._summarize_results()
        if failed_jobs > 0:
            result.merge(Result.fail(f"{failed_jobs} failed during this execution"))

        return result



"""
Tool for manipulating bundles for airgapped transfers.
"""
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process, hoppr_rerunner
from hoppr.exceptions import (
    HopprCredentialsError,
    HopprError,
    HopprExperimentalWarning,
    HopprLoadDataError,
    HopprPluginError,
)
from hoppr.logger import HopprLogger
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService, Credentials
from hoppr.models.manifest import Manifest
from hoppr.models.sbom import Component, ExternalReference, Property, Sbom
from hoppr.models.transfer import ComponentCoverage, Transfer
from hoppr.models.types import BomAccess, PurlType
from hoppr.result import Result

__version__ = "1.10.2"

__all__ = [
    "__version__",
    "BomAccess",
    "Component",
    "ComponentCoverage",
    "CredentialRequiredService",
    "Credentials",
    "ExternalReference",
    "hoppr_process",
    "hoppr_rerunner",
    "HopprContext",
    "HopprCredentialsError",
    "HopprError",
    "HopprExperimentalWarning",
    "HopprLoadDataError",
    "HopprLogger",
    "HopprPlugin",
    "HopprPluginError",
    "Manifest",
    "Property",
    "PurlType",
    "Result",
    "Sbom",
    "Transfer",
]



"""
CycloneDX data models
"""
from __future__ import annotations

import uuid

from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, ClassVar, Iterator, Literal, MutableMapping, TypeAlias, cast

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from pydantic import Extra, Field, root_validator, validator
from rapidfuzz import fuzz
from requests import HTTPError

import hoppr.net
import hoppr.utils

from hoppr.exceptions import HopprLoadDataError
from hoppr.models.affect import Affect
from hoppr.models.base import CycloneDXBaseModel, UniqueIDMap
from hoppr.models.licenses import LicenseChoice
from hoppr.models.types import LocalFile, OciFile, UrlFile

if TYPE_CHECKING:
    from packageurl import PackageURL

DictStrAny: TypeAlias = dict[str, Any]

__all__ = [
    "Component",
    "ExternalReference",
    "Hash",
    "Property",
    "Sbom",
]

SbomRef = Annotated[LocalFile | OciFile | UrlFile, Field(..., description="Reference to a local or remote SBOM file")]
SbomRefMap = Annotated[MutableMapping[SbomRef, "Sbom"], Field(default=...)]

FUZZY_MATCH_THRESHOLD = 85


def _extract_components(components: list[Component]) -> list[Component]:
    """
    Explicitly set scope of flattened components to `exclude`
    """
    for component in components:
        setattr(component, "scope", "excluded")

    return hoppr.utils.dedup_list(components)


def _extract_sbom_components(external_refs: list[ExternalReference]) -> list[Component]:
    """
    Extracts `external_refs` of type "bom" and returns the set of their components
    """
    components: list[Component] = []

    for ref in _get_bom_refs(external_refs):
        sbom = Sbom.load(_resolve_sbom_source(ref.dict()["url"]))
        sbom.components = _extract_components(sbom.components)
        components.extend(sbom.components)
        external_refs.remove(ref)

    return components


def _flatten_component(component: Component) -> list[Component]:
    """
    Helper function to flatten a component's subcomponents into a set
    """
    flattened = []

    for subcomponent in component.components or []:
        # Ensure validator is run to set `bom_ref`
        subcomponent = Component(**subcomponent.dict())
        setattr(subcomponent, "scope", "excluded")

        # Flatten nested components into top level components list
        flattened.append(subcomponent)

    component.components.clear()
    return flattened


def _get_bom_refs(external_refs: list[ExternalReference]) -> Iterator[ExternalReference]:
    """
    Get `externalReferences` of type "bom"
    """
    yield from (ref.copy(deep=True) for ref in (external_refs or []) if ref.type == "bom")


def _resolve_sbom_source(source: str) -> str | Path | DictStrAny:
    """
    Resolves an SBOM source as a file path, URL or `dict`
    """
    return Path(source.removeprefix("file://")).resolve() if source.startswith("file://") else source


def _validate_components(cls, components: list[Component]) -> list[Component]:
    """
    Validator to optionally flatten `components` list
    """
    if not cls.flatten:
        return components

    flattened = list(components)

    for component in components:
        flattened.extend(_flatten_component(component))

    return hoppr.utils.dedup_list(flattened)


def _validate_external_refs(cls, external_refs: list[ExternalReference], values: DictStrAny) -> list[ExternalReference]:
    """
    Validator to optionally resolve `externalReferences`
    """
    external_refs = [ExternalReference.create(ref) for ref in external_refs or []]

    if cls.deep_merge:
        external_ref_components = _extract_sbom_components(external_refs)
        values["components"] = hoppr.utils.dedup_list([*values.get("components", []), *external_ref_components])

    return external_refs


class ExternalReference(CycloneDXBaseModel, cdx.ExternalReference):
    """
    ExternalReference data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for ExternalReference model"
        extra = Extra.allow

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Hash(CycloneDXBaseModel, cdx.Hash):
    """
    Hash data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Hash model"
        extra = Extra.allow

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Property(CycloneDXBaseModel, cdx.Property):
    """
    Property data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Property model"
        extra = Extra.allow

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Component(CycloneDXBaseModel, cdx.Component):
    """
    Component data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Component model"
        extra = Extra.allow

    bom_ref: Annotated[str, Field(alias="bom-ref")] = None  # type: ignore[assignment]
    components: list[Component] = []  # type: ignore[assignment]
    externalReferences: list[ExternalReference] = []  # type: ignore[assignment]
    hashes: list[Hash] = []  # type: ignore[assignment]
    licenses: LicenseChoice = []  # type: ignore[assignment]
    properties: list[Property] = []  # type: ignore[assignment]

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}

    validate_components: classmethod = validator("components", allow_reuse=True, always=True)(_validate_components)
    validate_external_refs: classmethod = validator("externalReferences", allow_reuse=True)(_validate_external_refs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Component):
            return False

        if self.purl and other.purl:
            self_purl = hoppr.utils.get_package_url(self.purl)
            other_purl = hoppr.utils.get_package_url(other.purl)

            qual_keys = hoppr.utils.dedup_list({**self_purl.qualifiers, **other_purl.qualifiers})

            return all(
                [
                    self_purl.name == other_purl.name,
                    self_purl.type == other_purl.type,
                    self_purl.namespace == other_purl.namespace,
                    str(self_purl.version).removeprefix("v") == str(other_purl.version).removeprefix("v"),
                    self_purl.subpath == other_purl.subpath,
                    *[self._qualifier_match(key, self_purl, other_purl) for key in qual_keys],
                ]
            )

        if self.bom_ref and other.bom_ref:
            return self.bom_ref == other.bom_ref

        return False  # pragma: no cover

    @staticmethod
    def _qualifier_match(key: str, self_purl: PackageURL, other_purl: PackageURL) -> bool:
        # Compare only if both purls have a value for specified qualifier
        if key not in (self_qual := self_purl.qualifiers) or key not in (other_qual := other_purl.qualifiers):
            return True

        return fuzz.ratio(self_qual.get(key, ""), other_qual.get(key, "")) > FUZZY_MATCH_THRESHOLD

    @root_validator(allow_reuse=True, pre=True)
    @classmethod
    def validate_component(cls, values: DictStrAny) -> DictStrAny:
        """
        Validator to set a Component's `bom-ref` identifier if not set
        """
        if not any([values.get("purl"), values.get("bom-ref"), all([values.get("name"), values.get("version")])]):
            raise ValueError(
                "Either 'bom-ref' or 'purl' must be defined, or 'name' and 'version' must be defined on a component"
            )

        def _unescape_unicode(value: str) -> str:
            # Decode any unicode escape sequences, e.g. "\u0026" -> "&"
            return value.encode(encoding="utf-8").decode()

        if values.get("purl"):
            values["purl"] = _unescape_unicode(str(values["purl"]))

            if not values.get("version") and (version := hoppr.utils.get_package_url(values["purl"]).version):
                values["version"] = version

        bom_ref = str(
            values.get("purl")
            or f"{'@'.join(filter(None, [values.get('name'), values.get('version')]))}"
            or values.get("bom-ref")
        )

        values["bom-ref"] = _unescape_unicode(bom_ref)

        return values


Component.update_forward_refs()


class Vulnerability(CycloneDXBaseModel, cdx.Vulnerability):
    """
    Vulnerability data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Vulnerability model"
        extra = Extra.allow

    affects: Annotated[
        list[Affect],
        Field(description="The components or services that are affected by the vulnerability.", title="Affects"),
    ] = []  # type: ignore[assignment]

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Sbom(CycloneDXBaseModel, cdx.CyclonedxSoftwareBillOfMaterialsStandard):
    """
    Sbom data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Sbom model"

    components: list[Component] = []  # type: ignore[assignment]
    externalReferences: list[ExternalReference] = []  # type: ignore[assignment]
    specVersion: str = "1.5"
    version: int = 1
    vulnerabilities: list[Vulnerability] = []  # type: ignore[assignment]

    # Attributes not included in schema
    loaded_sboms: ClassVar[SbomRefMap] = {}
    unique_id_map: ClassVar[UniqueIDMap] = {}

    validate_components: classmethod = validator("components", allow_reuse=True, always=True)(_validate_components)
    validate_external_refs: classmethod = validator("externalReferences", allow_reuse=True)(_validate_external_refs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Sbom):
            return False

        return True if self.serialNumber == other.serialNumber else super().__eq__(other)

    @validator("metadata", allow_reuse=True, always=True)
    @classmethod
    def validate_metadata(cls, metadata: cdx.Metadata | None) -> cdx.Metadata:
        """
        Validator to populate and normalize `metadata` field
        """
        metadata = metadata or cdx.Metadata()

        # Generate `metadata` field containing Hoppr tool component
        hoppr_metadata: Any = CycloneDXBaseModel.create(cdx.Metadata(timestamp=datetime.now(timezone.utc)))
        hoppr_metadata.tools = CycloneDXBaseModel.create(cdx.ToolModel())
        hoppr_metadata.tools.components = [  # type: ignore[attr-defined]
            Component(
                type=cdx.Type("application"),
                name="hoppr",
                version=hoppr.__version__,
                bom_ref=f"pkg:pypi/hoppr@{hoppr.__version__}",
                purl=f"pkg:pypi/hoppr@{hoppr.__version__}",
                scope=cdx.Scope("excluded"),
            ),
        ]

        # Convert deprecated `list[Tool]` to `ToolModel` structure
        if isinstance(tools := metadata.tools, list):
            hoppr_metadata.tools.components.extend(  # type: ignore[attr-defined]
                hoppr.utils.dedup_list(
                    Component(
                        type=cdx.Type("application"),
                        name=str(tool.name),
                        version=tool.version,
                        hashes=[Hash.create(hash) for hash in tool.hashes or []],
                        externalReferences=[ExternalReference.create(ref) for ref in tool.externalReferences or []],
                        scope=cdx.Scope("excluded"),
                    )
                    for tool in tools
                )
            )

            metadata.tools = None

        # Merge any remaining metadata from the SBOM being validated into Hoppr metadata
        hoppr_metadata.merge(CycloneDXBaseModel.create(metadata or cdx.Metadata()))
        return cast(cdx.Metadata, hoppr_metadata)

    @root_validator(allow_reuse=True, pre=True)
    @classmethod
    def validate_sbom(cls, values: DictStrAny) -> DictStrAny:
        """
        Validator to standardize fields
        """
        values["$schema"] = "http://cyclonedx.org/schema/bom-1.5.schema.json"
        values["specVersion"] = "1.5"
        values["serialNumber"] = values.get("serialNumber", None) or uuid.uuid4().urn

        return values

    @classmethod
    def find_ref(cls, ref_type: Literal["local", "oci", "url"], location: str | Path) -> Sbom | None:
        """
        Look up SBOM object by reference

        Args:
            ref_type (Literal["local", "oci", "url"]): Type of SBOM reference
            location (str | Path): Location of SBOM reference

        Returns:
            Sbom | None: SBOM object if found, otherwise None
        """
        # pylint: disable=duplicate-code
        match ref_type:
            case "local":
                return cls.loaded_sboms.get(LocalFile(local=Path(location)), None)
            case "oci":
                return cls.loaded_sboms.get(OciFile(oci=str(location)), None)
            case "url":
                return cls.loaded_sboms.get(UrlFile(url=str(location)), None)
            case _:
                return None

    @classmethod
    def load(cls, source: str | Path | DictStrAny) -> Sbom:
        """
        Load SBOM from local file, URL, or dict
        """
        # pylint: disable=duplicate-code
        match source:
            case dict():
                sbom = cls(**source)
            case Path():
                # Convert source to relative path if in current working directory subpath
                source = source.resolve()
                source = source.relative_to(Path.cwd()) if source.is_relative_to(Path.cwd()) else source

                sbom = cls.parse_file(source)
                cls.loaded_sboms[LocalFile(local=source)] = sbom
            case str():
                try:
                    sbom_dict = hoppr.net.load_url(source)
                    if not isinstance(sbom_dict, dict):
                        raise TypeError("URL SBOM was not loaded as dictionary")

                    sbom = cls.parse_obj(sbom_dict)
                    url_ref = UrlFile(url=source)
                    cls.loaded_sboms[url_ref] = sbom
                except (HopprLoadDataError, HTTPError) as ex:
                    raise HopprLoadDataError from ex

        return sbom



"""
Supporting class for oras bundle plugin
"""
import json
import os
import platform

from datetime import datetime
from http import cookiejar

import jsonschema
import oras.defaults
import oras.oci
import oras.provider
import oras.schemas
import requests

from oras.container import Container
from oras.decorator import ensure_container

import hoppr


class Registry(oras.provider.Registry):
    """Override the default Oras Registry Class"""

    @ensure_container
    def push(self, container: Container, archives: list, logger):  # pylint: disable=W0221
        """
        Given a list of layer metadata (paths and corresponding mediaType) push.
        """
        # Prepare a new manifest
        manifest = oras.oci.NewManifest()
        self.session.cookies.set_policy(BlockAll())

        self.upload_blobs(container, archives, manifest, logger)

        # Prepare manifest and config (add your custom annotations here)
        manifest["annotations"] = {
            "org.opencontainers.image.created": str(datetime.now()),
            "ArtifactType": "Oras OCI Bundle",
            "Documentation": "https://oras.land/cli/",
        }
        config_file = "/tmp/oras-conf.json"
        config_blob = {"os": str(platform.platform()), "HopctlVersion": hoppr.__version__}
        self.setup_default_config(config_blob, config_file)
        conf, config_file = oras.oci.ManifestConfig(
            path=config_file, media_type="application/vnd.oci.image.config.v1+json"
        )
        conf["annotations"] = config_blob

        # Config is just another layer blob!
        logger.info(f"Uploading config to {container.uri}")
        logger.info(f"Config {conf}")
        logger.flush()
        response = self.upload_blob(config_file, container, conf)
        self._check_200_response(response)
        os.remove(config_file)

        # Final upload of the manifest
        manifest["config"] = conf
        # Try updating the manifest
        response = self.upload_manifest(manifest=manifest, container=container)
        self._check_200_response(response=response)
        return response

    def upload_blobs(self, container: Container, archives: list, manifest: dict, logger) -> dict:
        """
        Upload individual layers to oci registry

        Args:
            container (Container): oras container
            archives (list): list of layers to upload
            logger (_type_): hoppr logger
        """
        # Upload files as blobs
        for item in archives:
            blob = item.get("path")
            media_type = item.get("media_type")
            annots = item.get("annotations") or {}

            if not blob or not os.path.exists(blob):
                logger.info(f"Path {blob} does not exist or is not defined.")
                continue

            # Artifact title is basename or user defined
            blob_name = item.get("title") or os.path.basename(blob)

            # Create a new layer from the blob
            layer = oras.oci.NewLayer(blob, media_type, is_dir=False)
            logger.info(f"Preparing layer {layer}")

            # Update annotations with title we will need for extraction
            annots.update({oras.defaults.annotation_title: blob_name})
            layer["annotations"] = annots

            # update the manifest with the new layer
            manifest["layers"].append(layer)

            # Upload the blob layer
            logger.info(f"Uploading {blob} to {container.uri}")
            logger.flush()
            response = self.upload_blob(blob, container, layer)
            self._check_200_response(response)

        return manifest

    def upload_manifest(self, manifest: dict, container: Container) -> requests.Response:
        """
        Read a manifest file and upload it.

        Args:
            manifest (dict): Manifest to upload
            container (Container): Parsed container URI

        Returns:
            requests.Response: Upload response
        """
        self.reset_basic_auth()  # Should we really be reseting?
        jsonschema.validate(manifest, schema=oras.schemas.manifest)

        headers = {
            "Content-Type": oras.defaults.default_manifest_media_type,
            "Content-Length": str(len(manifest)),
        }

        hostname, *_ = container.registry.split(":")
        put_url = f"{self.prefix}://{hostname}/v2/{container.api_prefix}/manifests/{container.tag}"

        return self.do_request(put_url, "PUT", headers=headers, json=manifest)

    def setup_default_config(self, config_blob: dict, config_file: str):
        """
        Setup defualt oras object config
        """
        with open(config_file, "w", encoding="utf-8") as fout:
            json.dump(config_blob, fout, indent=4)


class BlockAll(cookiejar.CookiePolicy):
    """
    Block all cookies

    Args:
        cookiejar (CookiePolicy): _description_
    """

    def _block_all(self, *_, **__) -> bool:  # pragma: no cover
        return False

    netscape = True
    return_ok = set_ok = domain_return_ok = path_return_ok = _block_all
    rfc2965 = hide_cookie2 = False



"""
Collector plugin for NuGet packages
"""
from __future__ import annotations
from pydantic import SecretStr

import requests

from requests.auth import HTTPBasicAuth

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.result import Result


class CollectNugetPlugin(SerialCollectorPlugin):
    """
    Collector plugin for NuGet packages
    """

    supported_purl_types: list[str] = ["nuget"]
    products: list[str] = ["nuget/*"]
    system_repositories: list[str] = ["https://www.nuget.org/api/v2/package"]
    process_timeout = 60

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.manifest_repos: list[str] = []
        self.password_list: list[str] = []

    def get_version(self):
        return __version__

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        purl = hoppr.utils.get_package_url(comp.purl)
        self.get_logger().info(msg=f"Copying Nuget package from {comp.purl}", indent_level=2)

        nuget_url = "/".join([repo_url.rstrip("/"), purl.name, purl.version])

        authentication = None
        if creds is not None and isinstance(creds.password, SecretStr):
            authentication = HTTPBasicAuth(creds.username, creds.password.get_secret_value())
        response = requests.get(nuget_url, auth=authentication, timeout=120, proxies=self._get_proxies_for(repo_url))

        if response.status_code == 404:
            msg = (
                f"NuGet failed to locate package for {comp.purl}, "
                f"return_code={response.status_code}, "
                f"URL should be full domain with path up to package, "
                f"e.g. https://www.nuget.org/api/v2/package"
            )
            self.get_logger().debug(msg=msg, indent_level=2)

            return Result.fail(message=msg)

        if response.status_code not in range(200, 300):
            msg = f"NuGet failed to locate package for {comp.purl}, return_code={response.status_code}"
            self.get_logger().debug(msg=msg, indent_level=2)

            return Result.retry(message=msg)

        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)

        package_name = f"{'_'.join(filter(None, [purl.name, purl.version]))}.nupkg"

        package_out_path = target_dir / package_name

        package_out_path.write_bytes(response.content)

        self.set_collection_params(comp, repo_url, target_dir)
        return Result.success(return_obj=comp)
